"""
File Purpose: EppicSubsamplable
"""
import os
import re

# optional external modules
from ...tools import ImportFailed
try:
    import h5py
except ImportError as err:
    h5py = ImportFailed("h5py", err=err, locals=locals())

# internal imports
from .eppic_direct_loader import EppicDirectLoader
from ...dimensions import (
    Subsamplable,
)
from ...errors import SubsamplingError, SubsamplingFormatError
from ...tools import (
    UNSET,
)


''' --------------------- EppicSubsamplable --------------------- '''

class EppicSubsamplable(EppicDirectLoader, Subsamplable):
    '''manages subsampling of EPPIC output, and interpreting subsampled EPPIC output.

    self.subsample() outputs to {self.dirname}/subsampling_result:
        subsampled snapshots go to subsampling_result/parallel  (assuming full_read_mode='h5_2')
        subsampling info goes to subsampling_result/subsampling_info

        Users may prefer the simpler version if just looking for simple slicing of snaps/x/y/z:
            self.subsample_simple_slices() can easily serve this purpose.

    self.subsampling_info by default reads info from {self.dirname}/subsampling_info.
        If that directory doesn't exist, it will be None, and assume there is no subsampling.

    snap src objects for EPPIC will be Snap objects, not filepaths.
    '''

    # # # SUBSAMPLING IMPLEMENTATION DETAILS (REQUIRED BY PARENT) # # #

    def subsampling_snap_srcs(self):
        '''list of all snap srcs in self (before applying any subsampling).
        Here, just returns self.snaps.
        '''
        return self.snaps

    def snap_src_to_filepath(self, src):
        '''convert snap src to filepath. Returns self.snap_filepath(src).'''
        return self.snap_filepath(src)

    def rawvars_loadable(self, src):
        '''returns list of all directly loadble vars within snap src.
        Used by subsampling routines. Basically just calls self.directly_loadable_vars.
        '''
        return self.directly_loadable_vars(snap=src)

    def rawvar_load(self, var, src):
        '''load a single raw var from snap src.
        Used by subsampling routines. Basically just calls self.load_fromfile.
        '''
        return self.load_fromfile(var, snap=src)

    def rawvar_save(self, var, data, dst):
        '''saves a single raw var to snap dst.
        crash if this would require overwriting any existing data at dst.
        '''
        with h5py.File(dst, 'a') as f:
            if var in f:
                raise SubsamplingError(f'var {var!r} already exists in dst={dst!r}.')
            f[var] = data

    def rawvar_load_and_subsample(self, var, src, info):
        '''returns var loaded and subsampled from src.'''
        applier = info.get_applier(var)
        if applier.mode == 'slice' and self.full_read_mode == 'h5_2':
            # [EFF] slice before loading full array.
            path = self.snap_filepath(src)
            if not os.path.exists(path):
                raise SubsamplingError(f'file {path!r} does not exist.')
            with h5py.File(path, 'r') as f:
                data_obj = f[var]
                subsampled = applier.apply(data_obj)
            return subsampled
        else: # less efficient but more generic: subsample after loading full array.
            return super().rawvar_load_and_subsample(var, src, info)


    # # # SUBSAMPLING CONVENIENT METHODS # # #

    def subsample_maindims_simple(self, *, snap=None, x=None, y=None, z=None, **kw_subsample):
        '''self.subsample(subsampling_info corresponding to slicing snaps and maindims, only).

        snap, x, y, z: None or slice
            slice for this dim.
        '''
        info = self.subsampling_info_maindims_simple(snap=snap, x=x, y=y, z=z, as_json=False)
        return self.subsample(info, **kw_subsample)

    def subsampling_info_maindims_simple(self, *, snap=None, x=None, y=None, z=None, as_json=False):
        '''SubsamplingInfo corresponding to slicing snaps and maindims, only.
        
        snap, x, y, z: None or slice
            slice for this dim.
        as_json: bool
            if True, return json for info instead of converting to SubsamplingInfo object.
            (Could be useful if you want to use this json as a starting point then make small edits)
        '''
        result = {}
        if snap is not None:
            assert isinstance(snap, slice)
            result['forall'] = {'snaps': ['slice', {'snap': [snap.start, snap.stop, snap.step]}]}
        if not all(d is None for d in (x, y, z)):
            vdict = {'array_dims': list(self.maindims)}
            xyzslices = {}
            if x is not None: xyzslices['x'] = [x.start, x.stop, x.step]
            if y is not None: xyzslices['y'] = [y.start, y.stop, y.step]
            if z is not None: xyzslices['z'] = [z.start, z.stop, z.step]
            vdict['regex'] = {self._maindims_vars_regex: ['slice', xyzslices]}
            result['byvars'] = [vdict]
        return result if as_json else self.subsampling_info_cls(result)

    def subsample_maindims_at_snaps(self, *, x=None, y=None, z=None,
                                den=UNSET, flux=UNSET, nvsqr=UNSET, phi=UNSET, others=None,
                                **kw_subsample):
        '''subsample same maindims but different snaps for different vars.

        x, y, z: None or slice
            slice for this main dim.
        den, flux, nvsqr, phi: UNSET, None or slice
            slice for snaps for this variable.
            None --> do NOT slice snaps for this variable (but still slice maindims!)
        others: None or slice
            slice snaps for whichever of den, flux, nvsqr, and phi are UNSET.
            None --> do not slice snaps (but still slice maindims!)
        '''
        info = self.subsampling_info_maindims_at_snaps(x=x, y=y, z=z,
                                    den=den, flux=flux, nvsqr=nvsqr, phi=phi, others=others,
                                    as_json=False)
        return self.subsample(info, **kw_subsample)

    def subsampling_info_maindims_at_snaps(self, *, x=None, y=None, z=None,
                                den=UNSET, flux=UNSET, nvsqr=UNSET, phi=UNSET, others=None,
                                as_json=False):
        '''SubsamplingInfo corresponding to subsampling maindims at different snaps.

        x, y, z: None or slice
            slice for this main dim. All maindims variables will be sliced according to this.
        den, flux, nvsqr, phi: UNSET, None or slice
            slice for snaps for this variable.
            None --> do not slice snaps for this variable (but still slice maindims!)
        others: None or slice
            slice snaps for whichever of den, flux, nvsqr, and phi are UNSET.
            None --> do not slice snaps (but still slice maindims!)
        as_json: bool
            if True, return json for info instead of converting to SubsamplingInfo object.
            (Could be useful if you want to use this json as a starting point then make small edits)
        '''
        result = {}
        xyz_info = []
        if all(d is None for d in (x,y,z)):
            xyz_info = ['identity', None]
        else:
            xyz_slicer = {}
            if x is not None: xyz_slicer['x'] = [x.start, x.stop, x.step]
            if y is not None: xyz_slicer['y'] = [y.start, y.stop, y.step]
            if z is not None: xyz_slicer['z'] = [z.start, z.stop, z.step]
            xyz_info = ['slice', xyz_slicer]
        result['byvars'] = []
        VAR_TO_REGEX = dict(den='den[0-9]+', flux='flux[xyz][0-9]+', nvsqr='nvsqr[xyz][0-9]+', phi='phi')
        for var, val in dict(den=den, flux=flux, nvsqr=nvsqr, phi=phi).items():
            if val is UNSET:
                continue  # handled by "others"
            reg = VAR_TO_REGEX.pop(var)
            if val is None:
                if xyz_info[0] != 'identity':  # slice maindims but not snaps
                    vdict = {
                        'array_dims': list(self.maindims),
                        'regex': {reg: xyz_info}
                    }
                else:  # nothing to slice for this var
                    vdict = None
            else:  # slice snaps (and possibly also maindims)
                vdict = {
                        'array_dims': list(self.maindims),
                        'snaps': ['slice', {'snap': [val.start, val.stop, val.step]}],
                        'regex': {reg: xyz_info}
                    }
            # add vdict to result
            if vdict is not None:
                result['byvars'].append(vdict)
        if len(VAR_TO_REGEX) != 0:
            reg = '|'.join(VAR_TO_REGEX.values())
            if others is None:
                if xyz_info[0] != 'identity':  # slice maindims but not snaps
                    vdict = {
                        'array_dims': list(self.maindims),
                        'regex': {reg: xyz_info}
                    }
                else:  # nothing to slice for this var
                    vdict = None
            else:
                vdict = {
                        'array_dims': list(self.maindims),
                        'snaps': ['slice', {'snap': [others.start, others.stop, others.step]}],
                        'regex': {reg: xyz_info}
                    }
            # add vdict to result
            if vdict is not None:
                result['byvars'].append(vdict)
        return result if as_json else self.subsampling_info_cls(result)


    # # # READING SUBSAMPLED DATA -- HELPER METHODS # # #

    _maindims_vars_regex = '((den|flux[xyz]|nvsqr[xyz])[0-9]+)|phi'
    _vdist_vars_regex = 'vdist[0-9]+'

    def _apply_subsampling_to_maindims_coords(self, coords):
        '''apply subsampling to maindims coords.
        coords is a dict of {'x': xcoords, 'y': ycoords} possibly also 'z': zcoords.

        [TODO] check that vars actually all have the same maindims, else, crash?
            implementation now just assumes 'den0' exists and has same maindims as other vars.
        '''
        info = self.subsampling_info
        if info is None:
            return coords
        var = 'den0'  # assume this var exists and has same maindims as all others...
        return info.get_applier(var).apply1d(coords)

    def _apply_subsampling_to_vdist_coords(self, vcoords, fluid):
        '''apply subsampling to vdist coords for this fluid (a single Fluid)'''
        info = self.subsampling_info
        if info is None:
            return vcoords
        N = int(fluid)
        var = f'vdist{N}'
        return info.get_applier(var).apply1d(vcoords)

    def _get_subsampling_step(self, x, var='den0'):
        '''get subsampling step for var along this direction.
        Crash with SubsamplingFormatError if subsampling mode for var exists but is not 'slice'.
        '''
        info = self.subsampling_info
        if info is None:
            return 1
        applier = info.get_applier(var)
        if applier.mode != 'slice':
            raise SubsamplingFormatError(f'applier mode for var={var!r} is not "slice".')
        return applier.step(x)



"""
File Purpose: loading single-fluid MHD quantities related to Equation Of State (EOS).
"""

from .mhd_bases import MhdBasesLoader
from .mhd_er_tables import erTabInputManager
from ..defaults import DEFAULTS
from ..errors import (
    FormulaMissingError, LoadingNotImplementedError,
    InputError, InputMissingError,
)
from ..tools import simple_property, UNSET


''' --------------------- MhdEosLoader--------------------- '''

class MhdEosLoader(MhdBasesLoader):
    '''single-fluid MHD quantities related to Equation of State (EOS): ne, T, P.

    The implementation here assumes tables available at table=self.tabin[var],
        for var='ne', 'T', or 'P', and each having a table.interp(r=r, e=e) method,
        which gives value of var in 'raw' units.

    Future refactors of PlasmaCalcs might split the functionality here,
        e.g. if there are other types of tables, add code to manage table types.
    '''
    tabin_cls = erTabInputManager  # class for making tabin during self._default_tabin()

    tabin = simple_property('_tabin', setdefaultvia='_default_tabin',
        doc='''dict-like manager for Equation Of State tables; should include keys 'ne', 'T', 'P'.
        Each table=tabin[var] should have a table.interp(r=r, e=e) method,
            which returns value of var in 'raw' units, given values of r & e in 'raw' units.''')

    def _default_tabin(self):
        '''return default value of self.tabin: erTabInputManager(self.tabinputfile, u=self.u).'''
        tabinputfile = self.tabinputfile
        return self.tabin_cls(tabinputfile, u=self.u)

    tabinputfile = simple_property('_tabinputfile',
        doc='''path to tabinputfile, used by self._default_tabin() to create self.tabin.''')
    @tabinputfile.getter
    def tabinputfile(self):
        if hasattr(self, '_tabinputfile'):
            return self._tabinputfile
        else:
            errmsg = (f"{type(self).__name__}.tabinputfile not set. You might need to set it explicitly,\n"
                      f"or this subclass might still need to implement default value for tabinputfile.")
            raise InputMissingError(errmsg)


    # # # EOS MODE DISPATCH / CODE ARCHITECTURE # # #
    cls_behavior_attrs.register('eos_mode', default='table')
    EOS_MODE_OPTIONS = {
        'ideal': '''treat as ideal gas. P = n kB T = (gamma - 1) e, and can't get ne.''',
        'table': '''plug r and e into tables (see self.tabin) to get ne, T, P.'''
    }
    eos_mode = simple_property('_eos_mode', setdefaultvia='_default_eos_mode', validate_from='EOS_MODE_OPTIONS',
            doc='''mode for "Equation of State" related variables (ne, T, P).
            see EOS_MODE_OPTIONS for details about available options.''')

    def _default_eos_mode(self):
        '''default value for self.eos_mode. Here: 'table'. Subclass might override.'''
        return 'table'

    _EOS_MODE_TO_NE_VAR = {'ideal': 'nan', 'table': 'ne_fromtable'}
    @known_var(attr_deps=[('eos_mode', '_EOS_MODE_TO_NE_VAR')])
    def get_ne(self):
        '''electron number density. Depends on self.eos_mode; see self.EOS_MODE_OPTIONS for details.
        'ideal' --> cannot get ne. Crash with TypevarNanError.
        'table' --> ne from plugging r and e into EOS lookup tables (see self.tabin).
        [more options might be available (depending on subclass) -- check self.EOS_MODE_OPTIONS for details.]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        mode = self.eos_mode
        if mode not in self._EOS_MODE_TO_NE_VAR:
            raise LoadingNotImplementedError(f"{type(self).__name__}('ne') when eos_mode={mode!r}.")
        var = self._EOS_MODE_TO_NE_VAR[mode]
        if var == 'nan':
            self._handle_typevar_nan(errmsg=f"get_ne fails, when eos_mode={mode!r}.")
        return self(var)

    _EOS_MODE_TO_T_VAR = {'ideal': 'T_ideal', 'table': 'T_fromtable'}
    @known_var(attr_deps=[('eos_mode', '_EOS_MODE_TO_T_VAR')])
    def get_T(self):
        '''temperature. Depends on self.eos_mode; see self.EOS_MODE_OPTIONS for details.
        'ideal' --> T from ideal gas law: P_ideal = n kB T_ideal --> T_ideal = P_ideal / (n kB).
        'table' --> T from plugging r and e into EOS lookup tables (see self.tabin).
        [more options might be available (depending on subclass) -- check self.EOS_MODE_OPTIONS for details.]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        mode = self.eos_mode
        if mode not in self._EOS_MODE_TO_T_VAR:
            raise LoadingNotImplementedError(f"{type(self).__name__}('T') when eos_mode={mode!r}.")
        var = self._EOS_MODE_TO_T_VAR[mode]
        if var == 'nan':  # (no nans for T in MhdEosLoader, but subclass might want to use nan.)
            self._handle_typevar_nan(errmsg=f"get_T fails, when eos_mode={mode!r}.")
        return self(var)

    _EOS_MODE_TO_P_VAR = {'ideal': 'P_ideal', 'table': 'P_fromtable'}
    @known_var(attr_deps=[('eos_mode', '_EOS_MODE_TO_P_VAR')])
    def get_P(self):
        '''pressure. Depends on self.eos_mode; see self.EOS_MODE_OPTIONS for details.
        'ideal' --> P from ideal gas law: P_ideal = n kB T_ideal = (gamma - 1) e.
        'table' --> P from plugging r and e into EOS lookup tables (see self.tabin).
        [more options might be available (depending on subclass) -- check self.EOS_MODE_OPTIONS for details.]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        mode = self.eos_mode
        if mode not in self._EOS_MODE_TO_P_VAR:
            raise LoadingNotImplementedError(f"{type(self).__name__}('P') when eos_mode={mode!r}.")
        var = self._EOS_MODE_TO_P_VAR[mode]
        if var == 'nan':  # (no nans for P in MhdEosLoader, but subclass might want to use nan.)
            self._handle_typevar_nan(errmsg=f"get_P fails, when eos_mode={mode!r}.")
        return self(var)


    # # # EOS == IDEAL GAS # # #

    @known_var(deps=['e', 'gamma'])
    def get_P_ideal(self):
        '''pressure (from ideal gas law?) P = (gamma - 1) * e
        [TODO] when is this relation actually true? is it ideal gas law, or something else?
        '''
        return (self('gamma') - 1) * self('e')

    @known_var(deps=['P_ideal', 'n'])
    def get_T_ideal(self):
        '''temperature, assuming ideal gas law. P = n kB T --> T = P / (n kB)'''
        return self('P_ideal') / (self('n') * self.u('kB'))


    # # # EOS == TABLES # # #

    def _get_ertab_var_raw(self, var):
        '''get var in 'raw' units, from the eos tables, using single-fluid r and e from self.
        CAUTION: array values use [raw], but coords use [self.coords_units].
        see self.tabin.keys() for var options. gets value via interpolation.
        '''
        table = self.tabin[var]  # <-- in its own line to help with debugging in case of crash.
        with self.using(coords_units=self.coords_units_explicit, units='raw'):
            # table values are always in 'raw' units, but result coords are in self.coords_units.
            e = self('SF_e')  # 'SF' - value for SINGLE_FLUID mode.
            r = self('SF_r')
        return table.interp(r=r, e=e)  # [raw] units for values, [self.units] for coords.

    def get_ertab_var(self, var, ustr):
        '''get var in self.units units from the eos tables, using r and e from self.
        see self.tabin.keys() for var options. gets value via interpolation.
        ustr: str
            convert result from raw to self.units by multiplying by self.u(ustr).
        '''
        result = self._get_ertab_var_raw(var) * self.u(ustr)
        return self.record_units(result)

    @known_var(deps=['SF_e', 'SF_r'], aliases=['ne_tab'])
    def get_ne_fromtable(self):
        '''electron number density, from plugging r and e into eos tables (see self.tabin).'''
        return self.get_ertab_var('ne', 'n')

    @known_var(deps=['SF_e', 'SF_r'], aliases=['T_tab'])
    def get_T_fromtable(self):
        '''temperature, from plugging r and e into eos tables (see self.tabin).'''
        # note: multifluid T_fromtable assumes same T for all fluids.
        return self.get_ertab_var('T', 'temperature')

    @known_var(deps=['SF_e', 'SF_r'], aliases=['P_tab'])
    def get_P_fromtable(self):
        '''pressure, from plugging r and e into eos tables (see self.tabin).'''
        self.assert_single_fluid_mode('P')
        return self.get_ertab_var('P', 'pressure')



"""
File Purpose: collisional cross sections calculations
"""

from .cross_section_tools import CrossMapping
from ..quantity_loader import QuantityLoader

from ...errors import InputError, FluidKeyError
from ...tools import simple_property


class CollisionsCrossSectionsLoader(QuantityLoader):
    '''cross section calculations. Mapping stored in self.collcross_map.

    This subclass is especially not intended for direct use; see CollisionsLoader instead.
    e.g. some methods here assume existence of helper methods implemented in CollisionsLoader.
    '''
    cross_mapping_type = CrossMapping

    cls_behavior_attrs.register('collisions_cross_mapping', default=CrossMapping())
    collisions_cross_mapping = simple_property('_collisions_cross_mapping', setdefaultvia='_default_cross_mapping',
        doc='''the collisions cross sections mapping to use.
        keys are (fluid1, fluid2) pairs; values are CrossTable objects.
            see self.CROSS_TABLE_DEFAULTS for shorthand options for values.
            for convenient methods to build the cross mapping, see:
                self.set_collisions_crosstab, self.set_collisions_crosstab_defaults.''')

    def _default_cross_mapping(self):
        '''return default CrossMapping object to use as self.collisions_cross_mapping.
        Called when accessing self.collisions_cross_mapping when no value has been set yet.
        '''
        cls = self.cross_mapping_type
        result = cls()
        self.set_collisions_crosstab_defaults(crossmap=result)
        return result

    def set_collisions_crosstab(self, fluid1, fluid2, crosstab, *, crossmap=None):
        '''roughly, does self.collisions_cross_mapping[(fluid1, fluid2)] = crosstab,
        but this is a bit more convenient since it allows more shorthand options (see below)

        fluid1, fluid2: int, str, or Fluid
            the fluids to set the crosstab for.
            Fluid --> use the provided value.
            int or str --> infer the Fluid based on self.fluids.
        crosstab: None, str, or CrossTable
            None --> del self.collisions_cross_mapping[(fluid1, fluid2)], if possible.
            else --> self.collisions_cross_mapping[(fluid1, fluid2)] = crosstab.
                -- CrossTable --> will use the provided value.
                -- str --> will use the CrossTable corresponding to this filename or key,
                            key from self.CROSS_TABLE_DEFAULTS.
                        Note: self.collisions_cross_mapping.smart=False can disable this.
        crossmap: None or CrossMapping
            if provided, set value in crossmap instead of in self.collisions_cross_mapping.
        
        Example, these are all equivalent, if fluids[0]=='e' and fluids[1]=='H II':
            self.set_collisions_crosstab('e', 'H II', 'e-h')
            self.set_collisions_crosstab(0, 1, 'e-h-cross.txt')
            self.set_collisions_crosstab(self.fluids.get('e'), self.fluids.get('H II'),
                                        CrossTable.from_file('e-h-cross.txt'))
            self.collisions_cross_mapping[(self.fluids.get(0), self.fluids.get(1))] = 'e-h'
        '''
        f1 = self._as_single_fluid_or_jfluid(fluid1)
        f2 = self._as_single_jfluid_or_fluid(fluid2)
        if crossmap is None: crossmap = self.collisions_cross_mapping
        if crosstab is None:
            crossmap.pop((f1, f2), None)
        else:
            crossmap[(f1, f2)] = crosstab

    def get_collisions_crosstab(self, fluid1, fluid2, *, crossmap=None):
        '''roughly, returns self.collisions_cross_mapping[(fluid1, fluid2)],
        but, this is more convenient, because:
            - fluid1 and fluid2 can be provided in shorthand.
                can provide them as Fluid objects, ints, or strs.
        crossmap: None or CrossMapping
            if provided, get value from crossmap instead of from self.collisions_cross_mapping.
        '''
        f1 = self._as_single_fluid_or_jfluid(fluid1)
        f2 = self._as_single_jfluid_or_fluid(fluid2)
        if crossmap is None: crossmap = self.collisions_cross_mapping
        return crossmap[(f1, f2)]

    @known_var(load_across_dims=['fluid', 'jfluid'], aliases=['collcross'])
    def get_collisions_cross_section(self):
        '''cross section between self.fluid and self.jfluid.
        interpolates on self.collisions_cross_mapping based on mass-weighted temperature;
            cross_table = collisions_cross_mapping[(fluid, jfluid)]
            T_sj = (m_j * T_s + m_s * T_j) / (m_j + m_s)
            result_si = cross_table(T_sj, input='T', output='cross_si')
            result = result_si * self.u('area')  # convert to self.units unit system.

        if collisions_cross_mapping[(fluid, jfluid)] is not found:
            either return 0 (as an xarray), or raise QuantInfoMissingError,
            depending on self.collisions_mode. see help(type(self).collisions_mode) for details.
        '''
        fs = self.fluid
        fj = self.jfluid
        if fs == fj:  # exclude same-same collisions
            return self('0')  # 0 as xarray
        try:
            cross_table = self.collisions_cross_mapping[(fs, fj)]
        except KeyError:
            return self._handle_missing_collisions_crosstab()  # if crash, includes KeyError info in traceback.
        T_sj = self('T_sj')
        result_si = cross_table(T_sj, input='T', output='collcross_si')
        result = result_si * self.u('area', convert_from='si')
        return result

    # # # DEFAULTS # # #
    @property
    def CROSS_TABLE_DEFAULTS(self):
        '''dict of {shorthand: (filename, fc)} with shorthand useable in self.set_collisions_crosstab.'''
        return self.cross_mapping_type.cross_table_type.DEFAULTS

    def _set_collisions_crosstab_defaults_from(self, mode='bruno', *,
                                               e=None,
                                               H_I=None, H_II=None,
                                               He_I=None, He_II=None, He_III=None,
                                               crossmap=None,
                                               ):
        '''set as many CROSS_TABLE_DEFAULTS cross sections as possible, for provided fluids.
        see also: self.set_collisions_crosstab_defaults - it provides relevant fluids from self.

        e.g. if provided e, H_I, H_II, He_II, would set:
            crossmap[e, H_I] = 'e-h'
            crossmap[H_II, H_I] = 'h-p'
            crossmap[He_II, H_I] = 'h-he'
            crossmap[H_I, H_I] = 'h-h'
        if provided more fluids, would set more crossmaps too.
        see self.CROSS_TABLE_DEFAULTS for list of available defaults.

        mode: 'bruno' or 'vranjes'
            tells which defaults to use. (bruno is probably more accurate than vranjes.)
        e, H_I, H_II, He_I, He_II, He_III: None, Fluid, int, str, or other fluid specifier.
            specifiers for electrons, H(neutral), H+, He(neutral), He+, He++ fluids.
            None --> do not set default cross tables related to this fluid.
            else --> will find corresponding Fluid value in self.fluids or self.jfluids.
        crossmap: None or CrossMapping
            if provided, set values in crossmap instead of in self.collisions_cross_mapping.
        '''
        if crossmap is None: crossmap = self.collisions_cross_mapping
        # readability is more important than efficiency here!
        inputs = dict(e=e, H_I=H_I, H_II=H_II, He_I=He_I, He_II=He_II, He_III=He_III)
        p = {k: (v is not None) for k, v in inputs.items()}  # whether each value was provided.
        if mode == 'bruno':
            # -- bruno -- charged-neutral collisions --
            if p['e'] and p['H_I']:
                self.set_collisions_crosstab(e, H_I, 'e-h', crossmap=crossmap)
            if p['e'] and p['He_I']:
                self.set_collisions_crosstab(e, He_I, 'e-he', crossmap=crossmap)
            if p['H_I'] and p['H_II']:
                self.set_collisions_crosstab(H_II, H_I, 'h-p', crossmap=crossmap)
            if p['H_I'] and p['He_II']:
                self.set_collisions_crosstab(He_II, H_I, 'h-hep', crossmap=crossmap)
            if p['H_I'] and p['He_III']:
                self.set_collisions_crosstab(He_III, H_I, 'h-hepp', crossmap=crossmap)
            if p['He_I'] and p['H_II']:
                self.set_collisions_crosstab(H_II, He_I, 'he-p', crossmap=crossmap)
            if p['He_I'] and p['He_II']:
                self.set_collisions_crosstab(He_II, He_I, 'he-hep', crossmap=crossmap)
            if p['He_I'] and p['He_III']:
                self.set_collisions_crosstab(He_III, He_I, 'he-hepp', crossmap=crossmap)
            # -- bruno -- neutral-neutral collisions --
            if p['H_I']:
                self.set_collisions_crosstab(H_I, H_I, 'h-h', crossmap=crossmap)
            if p['H_I'] and p['He_II']:
                self.set_collisions_crosstab(He_II, H_I, 'he-h', crossmap=crossmap)
            if p['He_I']:
                self.set_collisions_crosstab(He_I, He_I, 'he-he', crossmap=crossmap)
        elif mode == 'vranjes':
            # -- vranjes -- charged-neutral collisions --
            if p['e'] and p['H_I']:
                self.set_collisions_crosstab(e, H_I, 'e-h_vranjes', crossmap=crossmap)
            if p['e'] and p['He_I']:
                self.set_collisions_crosstab(e, He_I, 'e-he_vranjes', crossmap=crossmap)
            if p['H_I'] and p['H_II']:
                self.set_collisions_crosstab(H_II, H_I, 'h-p_vranjes', crossmap=crossmap)
            if p['He_I'] and p['H_II']:
                self.set_collisions_crosstab(H_II, He_I, 'he-p_vranjes', crossmap=crossmap)
            # -- vranjes -- neutral-neutral collisions --
            if p['H_I']:
                self.set_collisions_crosstab(H_I, H_I, 'h-h_vranjes', crossmap=crossmap)
            if p['He_I']:
                self.set_collisions_crosstab(He_I, He_I, 'he-he_vranjes', crossmap=crossmap)
        else:
            raise InputError(f'mode={mode!r}; expected "bruno" or "vranjes".')

    def set_collisions_crosstab_defaults(self, mode='bruno', *, crossmap=None):
        '''set CROSS_TABLE_DEFAULTS cross sections for all relevant fluids in self.
        e.g. if self has fluids e, H_I, H_II, He_II, would set:
            crossmap[e, H_I] = 'e-h'
            crossmap[H_II, H_I] = 'h-p'
            crossmap[He_II, H_I] = 'h-he'
            crossmap[H_I, H_I] = 'h-h'
        if self has more relevant fluids, would set more crossmaps too.
        see self.CROSS_TABLE_DEFAULTS for list of available defaults.
        (note that order of keys doesn't matter for crossmap since it's a SymmetricPairMapping.)

        mode: 'bruno' or 'vranjes'
            tells which defaults to use. (bruno is probably more accurate than vranjes.)
        crossmap: None or CrossMapping
            if provided, set values in crossmap instead of in self.collisions_cross_mapping.

        gets relevant fluids from self.get_collisions_crosstab_default_fluids;
            subclass may wish to override that method.
        '''
        fluids = self.get_collisions_crosstab_default_fluids()
        self._set_collisions_crosstab_defaults_from(mode=mode, **fluids, crossmap=crossmap)

    _COLLISIONS_CROSSTAB_DEFAULT_FLUIDS_ALIASES = {
        'H_I'   : ['H_I',    'H I',    'H'],
        'H_II'  : ['H_II',   'H II',   'H+'],
        'He_I'  : ['He_I',   'He I',   'He'],
        'He_II' : ['He_II',  'He II',  'He+'],
        'He_III': ['He_III', 'He III', 'He++'],
    }

    def get_collisions_crosstab_default_fluids(self):
        '''return dict of fluids from self which have any corresponding CROSS_TABLE_DEFAULTS.
        keys will be e, H_I, H_II, He_I, He_II, He_III.
        values will be Fluid objects (from self.fluids or self.jfluids)

        checks multiple possible shorthand notations for each fluid:
            e     : fluids.get_electron(), jfluids.get_electron()
            H_I   : 'H_I',    'H I',    'H'
            H_II  : 'H_II',   'H II',   'H+'
            He_I  : 'He_I',   'He I',   'He'
            He_II : 'He_II',  'He II',  'He+'
            He_III: 'He_III', 'He III', 'He++'
        if multiple matches found in any case, raise FluidValueError.
            subclass could implement more sophisticated matching if needed;
            the implementation here aims to be flexible, but crash if anything is ambiguous.
        note: possible shorthands defined by self._COLLISIONS_CROSSTAB_DEFAULT_FLUIDS_ALIASES;
            subclass might override that instead of this method, if just needing to alter shorthands.
        '''
        result = {}
        # electrons
        try:
            result['e'] = self.fluids.get_electron()
        except FluidKeyError:  # no electron in self.fluids
            try:
                result['e'] = self.jfluids.get_electron()
            except FluidKeyError:  # no electron in self.jfluids
                pass
        # other fluids:
        for f, aliases in self._COLLISIONS_CROSSTAB_DEFAULT_FLUIDS_ALIASES.items():
            try:
                result[f] = self._get_fluid_or_jfluid_like(aliases)
            except FluidKeyError:  # this fluid not found in self.fluids or self.jfluids
                pass
                # (note, above will still crash with FluidValueError, if 2+ fluids found.)
        return result




# Configuration file for the Sphinx documentation builder.
# Started by copy-pasting from readthedocs' tutorial;
#   some of the stuff in here is probably not necessary.
#   Feel free to mess around with it.

import os
import sys

from PlasmaCalcs import __version__
from PlasmaCalcs import DocstringInfo

# Source code dir relative to this file;
# used by api.rst's autosummary
sys.path.insert(0, os.path.abspath(os.path.join('..', '..')))


# -- Project information

project = 'PlasmaCalcs'
release = __version__
version = __version__
copyright = '2025, PlasmaCalcs Developers'


# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

#html_theme = "pydata_sphinx_theme"  # <-- compiles >4x slower than sphinx_rtd_theme.
html_theme = 'sphinx_rtd_theme'


''' --------------------- reformat pc docstrings --------------------- '''

def reformat_pc_docstrings(app, what, name, obj, options, lines0):
    '''modify docstring formatting to be decent even if not written to respect rst standards.
    Many PlasmaCalcs docstrings don't respect sphinx / python standards,
        preferring instead to be more readable directly in the source code.
        (Also, PlasmaCalcs didn't have online docs set up for its first ~2 years,
            so nobody was checking if it respected sphinx standards...)
    Thankfully, sphinx provides "autodoc-process-docstring" hook to modify docstrings.
    That's when this function gets called (see sphinx docs on that hook for more details).

    app: the sphinx application object
    what: type of obj: 'module', 'class', 'function', 'method', 'attribute', 'exception'
    name: fully qualified name of obj
    obj: the obj itself
    options: options given to the sphinx directive
    lines: lines of the docstring. Edit in-place to modify the docstring.

    E.g. we want functions like this to not look ugly when rendered by sphinx:
        def f(x,y,z,t, **kw):
            """one line summary.
            Longer description, but we don't want to destroy the line breaks;
                we also want to keep any indents like this one.

            x: int. description about x
            y: None or any value
                description about y
                extends to multiple lines
            z: str   # has no description here
            t: bool.
                description about t
                extends to multiple lines
                    and sometimes includes sub-indents on those lines!
            additional kwargs go to ...

            returns something.
            """
            ...  # code for f goes here.

    [TODO] spend more time fiddling with this function, to make into more "official" format:
        e.g. :param p: for params, :returns: for return info...
    '''
    if len(lines0)<=1 or all(len(l.strip())==0 for l in lines0):
        return  # don't mess with anything if only 1 line, or empty.
    else:
        di = DocstringInfo('\n'.join(lines0))
        newlines = di.to_sphinx_lines()
        lines0[:] = newlines

def setup(app):
    app.connect("autodoc-process-docstring", reformat_pc_docstrings)



"""
File Purpose: tools related to using a supercomputer / computer cluster.
"""
import os
import re
import warnings

from .io_tools import attempt_literal_eval
from .iterables import DictWithAliases
from .os_tools import find_files_re
from .sentinels import RESULT_MISSING
from ..errors import FileContentsConflictError


''' --------------------- Finding jobfiles --------------------- '''

def find_jobfiles(dir=os.curdir, *, exclude=[]):
    '''find all jobfiles in this directory and all subdirectories.
    jobfiles are files which end with '.o' or '.oN' where N is an integer with any number of digits.

    exclude: str, or list of strs
        exclude any subdirectories whose name equals one of these strings or re.fullmatch one of these strings.
        E.g. exclude='*[.]io' will exclude all subdirectories whose name ends with '.io';
        exclude='parallel' will exclude all subdirectories whose name equals 'parallel'.

    returns list of abspaths to all jobfiles found.
    '''
    return find_files_re('.*[.]o[0-9]*', dir=dir, exclude=exclude)


''' --------------------- Handling slurm files --------------------- '''

def find_slurmfiles(dir=os.curdir, *, exclude=[]):
    '''find all slurmfiles in this directory and all subdirectories.
    slurmfiles are files which end with '.slurm'.

    exclude: str, or list of strs
        exclude any subdirectories whose name equals one of these strings or re.fullmatch one of these strings.
        E.g. exclude='*[.]io' will exclude all subdirectories whose name ends with '.io';
        exclude='parallel' will exclude all subdirectories whose name equals 'parallel'.

    returns list of abspaths to all slurmfiles found.
    '''
    return find_files_re('.*[.]slurm', dir=dir, exclude=exclude)

class SlurmOptionsDict(DictWithAliases):
    '''DictWithAliases of slurm-related options.
    DEFAULT_ALIASES in this class tells all default aliases.
    '''
    DEFAULT_ALIASES = {
        '-J': '--job-name',
        '-o': '--output',
        '-p': '--partition',
        '-N': '--nodes',
        '-n': '--total-tasks',
        '--ntasks-per-node': '--tasks-per-node',
        '-t': '--time',
        '-A': '--account',
    }


def read_slurm_options(filename, *, eval=True):
    '''returns SlurmOptionsDict of all slurm-related options defined in this file.
    options are defined as lines which look like one of the following:
        "#SBATCH -N 4"  --> key='-N', value=4.
        "#SBATCH --time=1:00:00"  --> key='--time', value='1:00:00'

    eval: bool
        whether to attempt ast.literal_eval(value) for each value. (False --> keep all as str.)
    '''
    filename = os.path.abspath(filename)  # <-- makes error messages more verbose, if crash later.
    with open(filename, 'r') as file:
        lines = file.readlines()
    lines = [line.strip() for line in lines]
    pattern = re.compile(r'#SBATCH\s+(-.+?)(?:\s+|=)(.+?)\s*(#.*)?')
    #pattern = re.compile(r'#SBATCH\s+(-.+?\s+.+?|--.+?=.+?)(.+?)\s*(#.*)?')
    matches = [pattern.fullmatch(line) for line in lines]
    matches = [match for match in matches if match is not None]
    options = SlurmOptionsDict()
    for match in matches:
        key, value, comment = match.groups()
        if eval:
            value = attempt_literal_eval(value)
        if key in options:
            warnmsg = (f"multiple definitions of {key!r} in file {filename!r}"
                       "\n(read_slurm_options will keep only the last definition.)")
            warnings.warn(warnmsg)
        options[key] = value
    return options

def slurm_options_here(dir=os.curdir, *, conflicts='crash', only=None, eval=True):
    '''returns SlurmOptionsDict of all slurm-related options in files within this directory.
    searches all slurmfiles within this directory.

    conflicts: 'crash', 'drop'
        tells how to handle conflicts (key with different value or missing in different files).
        'crash' --> raise FileContentsConflictError if any conflicts.
        'drop' --> drop all keys with any sort of conflict.
    only: None, str, or list of str
        if provided, only include these keys in the result
        (and, only check these keys for conflicts!).
    eval: bool
        whether to attempt ast.literal_eval(value) for each value. (False --> keep all as str.)
    '''
    dir = os.path.abspath(dir)
    slurmfiles = find_slurmfiles(dir)
    if len(slurmfiles) == 0:
        raise FileNotFoundError(f'no slurmfiles found in directory {dir!r}')
    if isinstance(only, str):
        only = [only]
    results = []
    for f in slurmfiles:
        options = read_slurm_options(f, eval=eval)
        if only is not None:
            options = SlurmOptionsDict({key: options[key] for key in only if key in options})
        results.append(options)
    if len(results) > 1:
        # need to check for options conflicts.
        for key, val0 in list(results[0].items()):
            for opts in results[1:]:
                conflict = (val0 != opts.get(key, RESULT_MISSING))
                if conflict and conflicts == 'crash':
                    errmsg = (f'found multiple values for key {key!r} in slurmfiles {slurmfiles!r}\n'
                                f'Values found: {[opts.get(key, RESULT_MISSING) for opts in results]}')
                    raise FileContentsConflictError(errmsg)
                elif conflict and conflicts == 'drop':
                    for opts in results:
                        del opts[key]
    return results[0]

def slurm_option_here(key, dir=os.curdir, *, eval=True):
    '''return the value of the slurm-related option with this key, from slurm files within dir.
    searches all slurmfiles within this directory.
    Crashes if key not present with same exact value, in all slurmfiles.

    Recognizes some aliases for some keys; see SlurmOptionsDict.DEFAULT_ALIASES.
        (e.g. '-N' and '--nodes' are aliases of each other)

    key: str
        option or alias to option. Should include leading '-' or '--'.
    eval: bool
        whether to attempt ast.literal_eval(value) for each value. (False --> keep all as str.)

    Equivalent:
        slurm_options_here(dir, only=[key], ...)[key]
    '''
    return slurm_options_here(dir, only=key, conflicts='crash', eval=eval)[key]



"""
File Purpose: MuramDirectLoader
"""
import os

import numpy as np
import xarray as xr

from ...defaults import DEFAULTS
from ...errors import CacheNotApplicableError, SnapValueError
from ...quantities import DirectLoader
from ...tools import (
    simple_property,
    product,
)


''' --------------------- MuramDirectLoader --------------------- '''

class MuramDirectLoader(DirectLoader):
    '''manages loading data directly from bifrost output files.'''

    # # # FILE PATH INFO # # #
    @property
    def snapdir(self):
        '''directory containing the snapshot files.
        Here, gives self.dirname, because Muram outputs are stored at top-level of directory.
        '''
        return self.dirname

    def snap_filepath(self, snap=None):
        '''convert snap to full file path for this snap, i.e. to the Header.NNN file.'''
        snap = self._as_single_snap(snap)
        return snap.filepath

    # # # DIRECT LOADING-RELATED INFO # # #
    def directly_loadable_vars(self, snap=None):
        '''return tuple of directly loadable variables for this snap.

        snap: None, str, int, Snap, or iterable indicating multiple snaps.
            None --> use self.snap.
            multiple snaps --> return result if all snaps have same result, else crash.
        '''
        kw = dict() if snap is None else dict(snap=snap)
        with self.using(**kw):
            multiple_snaps = self.snap_is_iterable()
            if multiple_snaps:
                snaps = self.snap
            else:
                snaps = [self._as_single_snap(snap)]
        result = None
        for snap in snaps:
            # result for this snap
            if hasattr(snap, 'exists_for') and not snap.exists_for(self):
                result_here = np.nan
            else:
                result_here = snap.directly_loadable_vars()
            # compare against known result
            if result is None:
                result = result_here
            elif result != result_here:
                errmsg = 'directly_loadable_vars differ between snaps. Retry with single snap instead.'
                raise SnapValueError(errmsg)
            # else, continue
        return result

    @property
    def data_array_shape(self):
        '''shape of each array of data stored in files. ('N0', 'N1', 'N2') from self.params.
        Shape with (x, y, z) dimensions is some transposition of the result;
        transpose order stored in 'layout.order' file, see also self.params['order'].
        '''
        return (self.params['N0'], self.params['N1'], self.params['N2'])

    @property
    def data_array_size(self):
        '''array size (number of elements) of each data array. (from data_array_shape)'''
        return product(self.data_array_shape)

    data_array_dtype = simple_property('_data_array_dtype', default=np.dtype('float32'),
            doc='''numpy dtype of each data array. default is np.dtype('float32').
            Used by np.memmap during self.load_fromfile.''')

    @property
    def data_array_nbytes(self):
        '''number of bytes in each data array. (from data_array_size and data_array_dtype)'''
        return self.data_array_size * self.data_array_dtype.itemsize

    data_array_order = simple_property('_data_array_order', default='F',
            doc='''ordering of data array axes: 'F' (Fortran) or 'C' (C order). Default 'F'.''')

    # # # LOAD FROMFILE # # #
    _RESULT_PRIM_VAR_TO_FILEBASE = {
        'result_prim_r': 'result_prim_0',
        'result_prim_ux': 'result_prim_1',
        'result_prim_uy': 'result_prim_2',
        'result_prim_uz': 'result_prim_3',
        'result_prim_e': 'result_prim_4',
        'result_prim_bx': 'result_prim_5',
        'result_prim_by': 'result_prim_6',
        'result_prim_bz': 'result_prim_7',
    }

    def _var_for_load_fromfile(self, varname_internal):
        '''return var, suitably adjusted to pass into load_fromfile().
        Here, follow the steps:
            (1) if self._loading_component, append self.component.
                E.g. 'result_prim_b' turns into 'result_prim_bz' when loading 'z' component.
            (2) convert result_prim_... via self._RESULT_PRIM_VAR_TO_FILEBASE, if applicable.
                E.g. 'result_prim_bx' becomes 'result_prim_5'.
        '''
        result = varname_internal
        if getattr(self, '_loading_component', False):
            result = result + str(self.component)
        result = self._RESULT_PRIM_VAR_TO_FILEBASE.get(result, result)
        return result

    # # # LOAD FROMFILE # # #
    def load_fromfile(self, fromfile_var, *args__None, snap=None, **kw__None):
        '''return numpy array of fromfile_var, loaded directly from file.

        fromfile_var: str
            the name of the variable to read, adjusted appropriately for loading fromfile.
            E.g. use 'result_prim_5' not 'bx', to get magnetic field x-component.
            See also: self._var_for_load_fromfile().
        snap: None, str, int, or Snap
            the snapshot number to load. if None, use self.snap.
        '''
        snap = self._as_single_snap(snap)  # expect single snap when loading from file.
        if hasattr(snap, 'exists_for') and not snap.exists_for(self):
            result = xr.DataArray(self.snap_dim.NAN, attrs=dict(units='raw'))
            return self.assign_snap_coord(result)  # [TODO][EFF] assign coords when making array instead...
        filepath = os.path.join(snap.dirname, f'{fromfile_var}.{snap.s}')
        result = np.memmap(filepath,
                           mode='r',  # read-only; never alters existing files!
                           shape=self.data_array_shape,
                           dtype=self.data_array_dtype,
                           order=self.data_array_order,
                           )
        result = result.transpose(self.params['order'])  # transpose to match maindims order ('x', 'y', 'z').
        result = np.asarray(result)  # convert to numpy array.
        return result



"""
File Purpose: elements "in" single-fluid mhd calculators
mhd calculators can later infer element densities via abundances, ionization via saha.
"""

from ..defaults import DEFAULTS
from ..dimensions import Fluid, FluidList
from ..errors import FluidKeyError, InputConflictError
from ..tools import alias, elementwise_property


''' --------------------- Element & ElementList --------------------- '''

class Element(Fluid):
    '''Fluid corresponding to an element, including ions AND neutrals.

    CAUTION: some code assumes (without checking) that Element objects are immutable.
        Changing an Element after creating it could lead to unexpected behavior.
    
    name: None or str
        name of this element. None --> name unknown.
        defaults assume names will be title-case element symbol, e.g. 'H', 'He';
            using different names is allowed but will prevent default lookups.
    i: None or int
        index of this element (within an ElementList). None --> index unknown.
    m: None or number
        mass, in atomic mass units. E.g., ~1 for H.
    ionize_ev: None or number
        first ionization potential [eV]. E.g., ~13.6 for H.
    saha_g1g0: None or number
        ratio of g (degeneracy of states) for g1 (ions) to g0 (neutrals). E.g., 0.5 for H.
    abundance: None or number
        abundance of this element relative to H:
            A(elem) = 12 + log10(n(elem) / n(H)), n = number density.
        if not provided, trying to get abundance-related quantities will crash.

    internally, q=None always, because elements have no charge info.
    '''
    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'name', 'i', 'm', 'ionize_ev', 'saha_g1g0', 'abundance'}

    def __init__(self, name=None, i=None, *, m=None,
                 ionize_ev=None, saha_g1g0=None, abundance=None):
        super().__init__(name=name, i=i, m=m)
        self.ionize_ev = ionize_ev
        self.saha_g1g0 = saha_g1g0
        self.abundance = abundance

    # # # ABUNDANCE STUFF # # #
    def n_per_nH(self):
        '''return n(elem)/n(H) for this element. n = number density'''
        return 10**(self.abundance - 12)
    def r_per_nH(self):
        '''return r(elem)/n(H) for this element. r = mass density'''
        return self.m * self.n_per_nH()

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [repr(val) for val in [self.name, self.i] if val is not None]
        if self.m is not None:
            contents.append(f'm={self.m:.3f}')
        if self.q is not None:  # <-- expect q=None always, though...
            contents.append(f'q={self.q}')
        if self.ionize_ev is not None:
            contents.append(f'ionize_ev={self.ionize_ev:.1f}')
        if self.saha_g1g0 is not None:
            contents.append(f'saha_g1g0={self.saha_g1g0:.1f}')
        if self.abundance is not None:
            contents.append(f'abundance={self.abundance:.1f}')
        return f'{type(self).__name__}({", ".join(contents)})'


class ElementList(FluidList):
    '''List of Element objects.
    
    cls.DEFAULTS provides defaults for infotypes: 'm', 'ionize_ev', 'saha_g1g0', 'abundance'
        dict of {infotype: {defaultname: {elementname: value of this infotype}}}
        e.g. {'m': {'physical': {'H': 1, 'He': 4}},
              'ionize_ev': {'physical': {'H': 13.6, 'He': 24.6}},
              'saha_g1g0': {'physical': {'H': 0.5, 'He': 2.0}},
              'abundance': {'photospheric_v0': {'H': 12, 'He': 10.93},
                            'photospheric_v1': {'H': 12, 'He': 11.00}}
        the default for 'physical' for m and ionize_ev are in PlasmaCalcs.DEFAULTS.

    defaults assume names will be title-case element symbol, e.g. 'H', 'He';
        using different names is allowed but will prevent default lookups.
    '''
    value_type = Element

    DEFAULTS = {
        'm':         {'physical': DEFAULTS.PHYSICAL.M_AMU},
        'ionize_ev': {'physical': DEFAULTS.PHYSICAL.IONIZE_EV},
        'saha_g1g0': {'physical': DEFAULTS.PHYSICAL.SAHA_G1G0},
        'abundance': {},
        }

    ionize_ev = elementwise_property('ionize_ev')
    saha_g1g0 = elementwise_property('saha_g1g0')
    abundance = elementwise_property('abundance')

    # # # CREATION OPTIONS # # #
    @classmethod
    def from_names(cls, names=None, *names_as_args,
                   m='physical', ionize_ev='physical', saha_g1g0='physical', abundance={},
                   missing_ok=True, **common_info):
        '''create ElementList from names and dicts of info about each element.

        names: None or list of str
            names of elements.
            None --> use keys of other dicts.
            list --> use these names in this order.
            defaults assume names will be title-case element symbol, e.g. 'H', 'He';
                using different names is allowed but will prevent default lookups.
        names_as_args: additional names
            only allowed if names is a string; can list extra strings here.
            from_names('H', 'He', ...) equivalent to from_names(['H', 'He', ...]).
        m: str or dict
            masses of elements, in atomic mass units. E.g., {'H': 1}.
            str --> use cls.DEFAULTS['m'][m].
        ionize_ev: str or dict
            first ionization potentials of elements [eV]. E.g., {'H': 13.6}.
            str --> use cls.DEFAULTS['ionize_ev'][ionize_ev]
        saha_g1g0: str or dict
            ratio of g (degeneracy of states) for g1 (ions) to g0 (neutrals). E.g., {'H': 0.5}.
            str --> use cls.DEFAULTS['saha_g1g0'][saha_g1g0]
        abundance: str or dict
            abundances of elements relative to H. E.g., {'H': 12}.
            A(elem) = 12 + log10(n(elem) / n(H)), n = number density.
            str --> use cls.DEFAULTS['abundance'][abundance]
        missing_ok: bool
            whether to allow missing values for m, ionize_ev, or abundance.
            if False, missing values (i.e., None) will cause crash with FluidKeyError.
        additional kwargs passed to Element.__init__ for every element.
        '''
        # bookkeeping
        if isinstance(m, str):
            m = cls.DEFAULTS['m'][m]
        if isinstance(ionize_ev, str):
            ionize_ev = cls.DEFAULTS['ionize_ev'][ionize_ev]
        if isinstance(saha_g1g0, str):
            saha_g1g0 = cls.DEFAULTS['saha_g1g0'][saha_g1g0]
        if isinstance(abundance, str):
            abundance = cls.DEFAULTS['abundance'][abundance]
        if len(names_as_args) > 0:
            if not isinstance(names, str):
                raise InputConflictError('names_as_args only allowed if names is a string')
        if names is None:  # names = set union of all other keys, but maintain ordering.
            names = list(m.keys())
            names.extend(k for k in ionize_ev if k not in names)
            names.extend(k for k in abundance if k not in names)
        elif isinstance(names, str):
            names = [names]
            names.extend(names_as_args)
        # getting info for each Element
        infos = []
        for name in names:
            info = {
                'name': name,
                'm': m.get(name, None),
                'ionize_ev': ionize_ev.get(name, None),
                'saha_g1g0': saha_g1g0.get(name, None),
                'abundance': abundance.get(name, None),
            }
            if not missing_ok:
                for k in ('m', 'ionize_ev', 'saha_g1g0', 'abundance'):
                    if info[k] is None:
                        raise FluidKeyError(f'{k}.get({name!r}, None) is None, when missing_ok=False')
            infos.append(info)
        # create & return ElementList
        return cls.from_dicts(infos, **common_info)

    @classmethod
    def register_default(cls, infotype, defaultname, element2value):
        '''register a default lookup table for an infotype.
        e.g. cls.register_default('abundance', 'photospheric_v0', {'H': 12, 'He': 10.93})
        '''
        cls.DEFAULTS[infotype][defaultname] = element2value

    # [TODO] option to make ElementList with subset of elements but adjust abundances to get same total n.

    # # # ABUNDANCE STUFF # # #
    def n_per_nH(self, elem=None):
        '''n(elem)/n(H) for self.get(elem) (all elems in self if None).'''
        ee = self.get(elem)
        return 10**(ee.abundance - 12)
    def r_per_nH(self, elem=None):
        '''r(elem)/n(H) for self.get(elem) (all elems in self if None).'''
        ee = self.get(elem)
        return ee.m * ee.n_per_nH()

    def ntot_per_nH(self):
        '''sum(n(elem)/n(H)) == n(total)/n(H). Summed across all elems in self.'''
        return sum(self.n_per_nH())
    def rtot_per_nH(self):
        '''sum(r(elem)/n(H)) == r(total)/n(H). Summed across all elems in self.'''
        return sum(self.r_per_nH())

    def n_per_ntot(self, elem=None):
        '''n(elem)/n(total) for self.get(elem) (all elems in self if None).'''
        return self.n_per_nH(elem) / self.ntot_per_nH()
    def r_per_rtot(self, elem=None):
        '''r(elem)/r(total) for self.get(elem) (all elems in self if None).'''
        return self.r_per_nH(elem) / self.rtot_per_nH()

    def mtot(self):
        '''return "average" mass of all elements in self, weighted by n (implied by abundances).
        mtot = sum_x(mx ax) / sum_x(ax), where ax = nx / nH, and x is an element in self.

        Note: mtot = rtot / ntot. This can be proven as follows:
            rtot = sum_x(rx) = sum_x(mx nx) = sum_x(mx ax nH) = sum_x(mx ax) nH
            ntot = sum_x(nx) = sum_x(ax nH) = sum_x(ax) nH
            --> rtot / ntot = sum_x(mx ax) / sum_x(ax)
        (note: ntot does not include electrons)
        '''
        mx = self.m
        ax = self.n_per_nH()
        return sum(mx * ax) / sum(ax)



"""
File purpose: tools related to subplots.
"""
import builtins   # for unambiguous slice()

import numpy as np
import matplotlib.pyplot as plt

from .plot_settings import PlotSettings, PlotSettingsMixin
from .plot_tools import (
    colorbar, _paramdocs_colorbar, find_mappable,
    maintaining_current_axes,
    set_min_n_ticks,
)
from ..tools import (
    simple_property,
    format_docstring,
    UNSET,
)
from ..errors import (
    InputError, InputConflictError,
    MappableNotFoundError, PlottingAmbiguityError, TooManySubplotsError,
)

from ..defaults import DEFAULTS


''' --------------------- tools for subplots --------------------- '''

def subplots_figsize(nrows=1, ncols=1, axsize=None, *,
                     left=None, right=None, bottom=None, top=None, wspace=None, hspace=None):
    '''returns figsize to use for subplots so that every subplot has size axsize.
    axsize: None, number, or (width, height)
        size of a single subplot, in inches.
        if None, use DEFAULTS.PLOT.SUBPLOTS_AXSIZE.
        if number, use width=height=axsize.
    left, right, bottom, top, wspace, hspace: None or number (probably between 0 and 1)
        corresponding value that will be used during plt.subplots_adjust;
        figsize must be adjust accordingly to account for whitespace between and around subplots.
        None --> plt.rcParams['figure.subplot.left'], plt.rcParams['figure.subplot.right'], etc.
        left, right, bottom, top: tell position of edge of subplots grid, as fraction of figure width / height.
        wspace, hspace: tell width of padding between subplots, as fraction of axsize width / height.
    '''
    # [TODO] why do I still need to fidget with left/top/bottom to get supxlabel, supylabel, suptitle in good positions?
    # (note: it seems to be especially bad when doing nrow or ncol=1 and the other one being a large value (e.g. 5).)
    if axsize is None:
        axsize = DEFAULTS.PLOT.SUBPLOTS_AXSIZE
    try:
        axwidth, axheight = axsize
    except TypeError:
        axwidth = axheight = axsize
    if left is None: left = plt.rcParams['figure.subplot.left']
    if right is None: right = plt.rcParams['figure.subplot.right']
    assert right > left, "can't make plot when right < left!"
    if bottom is None: bottom = plt.rcParams['figure.subplot.bottom']
    if top is None: top = plt.rcParams['figure.subplot.top']
    assert top > bottom, "can't make plot when top < bottom!"
    if wspace is None: wspace = plt.rcParams['figure.subplot.wspace']
    if hspace is None: hspace = plt.rcParams['figure.subplot.hspace']
    # calculate figsize
    left_margin, right_margin = left, 1 - right
    subplots_width_fraction = 1 - left_margin - right_margin  # fraction of figure width remaining for the subplots
    subplots_width_inches = ((ncols-1)*wspace + ncols)*axwidth
    total_width = subplots_width_inches / subplots_width_fraction
    bottom_margin, top_margin = bottom, 1 - top
    subplots_height_fraction = 1 - bottom_margin - top_margin  # fraction of figure height remaining for the subplots
    subplots_height_inches = ((nrows-1)*hspace + nrows)*axheight
    total_height = subplots_height_inches / subplots_height_fraction
    return (total_width, total_height)
    

''' --------------------- Subplots --------------------- '''

@format_docstring(subplots_doc=plt.subplots.__doc__)
def subplots(nrows=1, ncols=1, *, squeeze=False, cls=None, **kw_subplots):
    '''returns a new Subplots object. Like calling plt.subplots(...), but a bit fancier.
    args & kwargs go to Subplots(...), then to plt.subplots(...).

    To avoid ambiguity, squeeze = True is not allowed.
        I.e., the axs will always be 2D, for the resulting Subplots object.

    cls: None or type
        the type to use for the Subplots class.
        None --> use DEFAULTS.PLOT.SUBPLOTS_TYPE
                (which is set inside subplots.py, and might be adjusted by other modules.
                it is probably Subplots or a subclass of Subplots.)

    For reference the docs for plt.subplots:
    ----------------------------------------
    {subplots_doc}
    '''
    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
    if cls is None:
        cls = DEFAULTS.PLOT.SUBPLOTS_TYPE
    return cls(nrows=nrows, ncols=ncols, squeeze=squeeze, **kw_subplots)

@PlotSettings.format_docstring(subplots_doc=plt.subplots.__doc__)
class Subplots(PlotSettingsMixin):
    '''grid of subplots, with methods for plotting on them.

    axsize: {axsize}
    figsize: {figsize}
    {kw_subplots_adjust}
    {kw_share_axlike}
    {kw_share_ax}
    max_nrows_ncols: {max_nrows_ncols}
    squeeze: {squeeze}
    min_n_ticks: {min_n_ticks}

    fig: None or plt.Figure
        if provided, use fig.subplots(...) instead of plt.subplots(...)

    Notes:
        - axs is always 2D, even if nrows=1 or ncols=1.

    Additional kwargs go to plt.subplots.

    For reference, here are the docs from plt.subplots:
    ---------------------------------------------------
    {subplots_doc}
    '''
    axes_class = UNSET

    def __init__(self, nrows=1, ncols=1, *, sharexlike=UNSET, shareylike=UNSET, fig=None, **kw):
        # # plot_settings # #
        super().__init__(nrows=nrows, ncols=ncols, sharexlike=sharexlike, shareylike=shareylike, **kw)
        ps = self.plot_settings
        # max_nrows_ncols:
        max_nrows_ncols = ps['max_nrows_ncols']
        if nrows > max_nrows_ncols or ncols > max_nrows_ncols:
            errmsg = f'nrows={nrows} or ncols={ncols} exceeds the non-None max_nrows_ncols={max_nrows_ncols}.'
            raise TooManySubplotsError(errmsg)
        # squeeze:
        if ps['squeeze']:
            raise InputError('squeeze=True is not allowed for Subplots(...).')
        # sharexlike, shareylike, use same value as sharex, sharey if provided:
        if sharexlike is UNSET and ps['sharex'] is not None:
            ps['sharexlike'] = ps['sharex']
        if shareylike is UNSET and ps['sharey'] is not None:
            ps['shareylike'] = ps['sharey']
        # figsize:
        if ps.get('figsize', default=UNSET) is UNSET:
            subplots_figsize_keys = ('axsize', 'left', 'right', 'bottom', 'top', 'wspace', 'hspace')
            ps['figsize'] = self.subplots_figsize(nrows=nrows, ncols=ncols, **{key: ps[key] for key in subplots_figsize_keys})
        elif ps.get('axsize', default=UNSET) is not UNSET:
            raise InputConflictError('cannot provide both axsize and figsize.')
        # axes_class:
        if (self.axes_class is not None) and ps.get('axes_class', default=None) is None:
            ps['axes_class'] = self.axes_class
        # # make the subplots # #
        kw_fig = ps.get_mpl_kwargs('plt.figure')
        if fig is None:
            fig = plt.figure(**kw_fig)
        kw_subplots = ps.get_mpl_kwargs('fig.subplots')
        if DEFAULTS.DEBUG >= 4:
            print('using kw_subplots', kw_subplots)
        axs = fig.subplots(nrows=nrows, ncols=ncols, **kw_subplots)
        assert axs.ndim == 2, f'expected axs.ndim == 2, got {axs.ndim}. Probably made a coding error...'
        self.fig = fig
        self.axs = axs
        # # cleanup plot formatting # #
        self.subplots_adjust()
        self.remove_redundant_labels(which={x for x in ('x', 'y') if ps[f'share{x}like']==True})
        self.apply_misc_formatting()

    subplots_figsize = staticmethod(subplots_figsize)

    # # # AXS # # #
    axs = simple_property('_axs', doc='''array of axes objects.''')
    @axs.setter  # update axs setter so that setting axs updates self.cache_state too.
    def axs(self, value):
        '''set self._axs=value; also update caching state.'''
        self._axs = value
        self.cache_state += 1

    cache_state = simple_property('_cache_state', default=0, doc='''int, associated with cached results.''')

    shape = property(lambda self: self.axs.shape, doc='''shape of axs array, as (nrows, ncols).''')
    nrows = property(lambda self: self.axs.shape[0], doc='''number of rows in axs array.''')
    ncols = property(lambda self: self.axs.shape[1], doc='''number of columns in axs array.''')

    @property
    def ax_idx(self):
        '''array of indices (irow, icol), with same shape as axs array, dtype=object'''
        try:
            cached_state, cached_result = self._ax_idx
        except AttributeError:
            pass
        else:
            if cached_state == self.cache_state:
                return cached_result
        # else, not found in cache; calculate result
        result = np.empty(self.shape, dtype=object)
        for irow, row in enumerate(self.axs):
            for icol, _ax in enumerate(row):
                result[(irow, icol)] = (irow, icol)
        # store in cache then return
        self._ax_idx = (self.cache_state, result)
        return result

    def __getitem__(self, idx):
        '''return self.axs[idx]'''
        return self.axs[idx]

    def __len__(self):
        '''return len(self.axs)'''
        return len(self.axs)

    # # # ITERATING # # #
    def __iter__(self):
        '''when iterating, return (self.fig, self.axs). To iterate over axs, see self.iter_ax.'''
        return iter((self.fig, self.axs))

    # # # ITER_AX, AX_APPLY # # #
    _iter_ax_paramdocs = {
        'slice': '''None, int, slice, or tuple
            if provided, only include the axes from axs[slice].
            when in this mode, (irow, icol) will still correspond to self.axs[(irow, icol)],
                (not self.axs[slice][irow, icol]).''',
        'sca': '''bool
            if True, call plt.sca(ax) before yielding each ax.''',
        'restore': '''bool
            if True, restore original plt.gca() after iteration is stopped.''',
    }

    def sca(self, irow, icol):
        '''set current axis to self.axs[irow, icol]'''
        plt.sca(self.axs[irow, icol])

    @format_docstring(**_iter_ax_paramdocs)
    def iter_ax(self, slice=None, *, sca=True, restore=True, skip=None):
        '''iterate over axs, one at a time, yielding ((irow, icol), ax) for each ax.
        slice: {slice}
        sca: {sca}
        restore: {restore}
        skip: None or callable of ((irow, icol), ax) --> bool
            if provided, skip axes for which skip(irow, icol, ax) returns True.
        '''
        with maintaining_current_axes(enabled=restore):
            if slice is None:
                slice = builtins.slice(None)
            ax_idx = self.ax_idx[slice]
            for sliced_index, ax in np.ndenumerate(self.axs[slice]):
                ax_idx_here = ax_idx if isinstance(ax_idx, tuple) else ax_idx[sliced_index]
                if skip is None or not skip(ax_idx_here, ax):
                    if sca: plt.sca(ax)
                    yield (ax_idx_here, ax)

    @format_docstring(**_iter_ax_paramdocs)
    def iter_row(self, irow, *, sca=True, restore=True, **kw_iter_ax):
        '''iterate over axs in row irow, yielding ((irow, icol), ax) for each ax. See also: self.iter_ax
        irow: int or slice
            row index(es) to iterate over. Can be negative int, e.g. -1 will be the bottom row.
        {sca}
        {restore}
        '''
        return self.iter_ax(slice=(irow, slice(None)), sca=sca, restore=restore, **kw_iter_ax)

    @format_docstring(**_iter_ax_paramdocs)
    def iter_col(self, icol, *, sca=True, restore=True, **kw_iter_ax):
        '''iterate over axs in column icol, yielding ((irow, icol), ax) for each ax. See also: self.iter_ax
        icol: int or slice
            column index(es) to iterate over. Can be negative int, e.g. -1 will be the rightmost column.
        {sca}
        {restore}
        '''
        return self.iter_ax(slice=(slice(None), icol), sca=sca, restore=restore, **kw_iter_ax)

    def ax_apply(self, f, *, sca=True):
        '''return numpy array of f(ax) applied to each ax. result has dtype=object.
        if sca, call plt.sca() before working on each ax.
        '''
        result = np.empty(self.shape, dtype=object)
        for (irow, icol), ax in self.iter_ax(sca=sca):
            result[(irow, icol)] = f(ax)
        return result

    # # # LABELS # # #
    def xlabel(self, xlabel, mode=None, *, only=True, **kw):
        '''set xlabel on relevant axs, or as supxlabel.
        mode: None or str in ('edge', 'all', 'sup')
            None --> use self.xlabel_mode
            'edge' --> only set xlabel on axs in bottom row
            'all' --> set xlabel on all axs.
            'sup' --> set xlabel on self.fig, as supxlabel.
        only: bool, default True
            if only, then also set xlabel to '' on all other axes.
        kwargs are passed to ax.set_xlabel or fig.supxlabel, as appropriate.
        '''
        if mode is None:
            mode = 'edge' if self.plot_settings['sharexlike'] else 'all'
        if mode=='edge':
            for _, ax in self.iter_row(-1):
                ax.set_xlabel(xlabel, **kw)
            if only:
                for _, ax in self.iter_row(slice(0, -1)): ax.set_xlabel('')
        elif mode=='all':
            for _, ax in self.iter_ax():
                ax.set_xlabel(xlabel, **kw)
        elif mode=='sup':
            self.fig.supxlabel(xlabel, **kw)
            if only:
                for _, ax in self.iter_ax(): ax.set_xlabel('')
        else:
            raise InputError(f'invalid mode: {mode}')

    def ylabel(self, ylabel, mode=None, *, only=True, **kw):
        '''set ylabel on relevant axs, or as supylabel.
        mode: None or str in ('edge', 'all', 'sup')
            None --> use self.xlabel_mode
            'all' --> set ylabel on all axs.
            'sup' --> set ylabel on self.fig, as supylabel.
        only: bool, default True
            if only, then also set ylabel to '' on all other axes.
        kwargs are passed to ax.set_ylabel or fig.supylabel, as appropriate.
        '''
        if mode is None:
            mode = 'edge' if self.plot_settings['shareylike'] else 'all'
        if mode=='edge':
            for _, ax in self.iter_col(0):
                ax.set_ylabel(ylabel, **kw)
            if only:
                for _, ax in self.iter_col(slice(1, None)): ax.set_ylabel('')
        elif mode=='all':
            for _, ax in self.iter_ax():
                ax.set_ylabel(ylabel, **kw)
        elif mode=='sup':
            self.fig.supylabel(ylabel, **kw)
            if only:
                for _, ax in self.iter_ax(): ax.set_ylabel('')
        else:
            raise InputError(f'invalid mode: {mode}')

    def remove_redundant_labels(self, which=['x', 'y'], *, ignore_empty=True):
        '''remove labels which are redundant, e.g. ticklabels,
        ylabels except in left col, xlabels except in bottom row.
        if sharexlike=False or shareylike=False, will not remove the corresponding labels.

        if self.plot_settings['polar'], keep xticklabels (i.e. angles) on top row instead of bottom.

        which: iterable of str in ('x', 'y')
            which labels to remove. Default: ('x', 'y')
        ignore_empty: bool
            whether to ignore empty axes (checked via ax.has_data())
        '''
        for x in which:
            self._remove_redundant_ticklabels(x, ignore_empty=True)
            self._remove_redundant_labels(x, ignore_empty=True)

    def _remove_redundant_ticklabels(self, x, *, ignore_empty=True):
        '''remove x or y ticklabels in self which are redundant, i.e. same across col (if 'x') or row (if 'y').
        x: 'x' or 'y'.
        ignore_empty: bool, whether to ignore axes in this column which have no data.
        '''
        if x not in ('x', 'y'): raise InputError(f'invalid x: {x!r}, expected "x" or "y".')
        _override = self.plot_settings[f'sharexlike'] if x=='x' else self.plot_settings[f'shareylike']
        for iline in (range(self.ncols) if x=='x' else range(self.nrows)):
            if _override is UNSET:
                common = self._get_common_tick_info(x, iline, ignore_empty=ignore_empty)
            else:
                common = (True if _override else None)
            if common is not None:
                axes = self.axs[:, iline][::-1] if x=='x' else self.axs[iline, :]
                if x=='x' and self.plot_settings.get('polar', last_resort_default=False):
                    axes = axes[::-1]
                first_nonempty = True
                for ax in axes:
                    if ignore_empty and not ax.has_data():
                        continue
                    elif first_nonempty:  # don't remove labels from first nonempty ax.
                        first_nonempty = False
                    else:
                        ax.set_xticklabels([]) if x=='x' else ax.set_yticklabels([])

    def _get_common_tick_info(self, x, iline, *, ignore_empty=True):
        '''return the x or y tick_info shared by all nonempty plots in this line, or None if that's impossible.
        tick_info is (xticklocs, xticklabels, xlim) (or y if x='y'), with str xticklabels (not Text objects).
        x: 'x' or 'y'.
        iline: int, index of the line to check. (icol if 'x', irow if 'y')
        ignore_empty: bool, whether to ignore axes in this column which have no data.
        '''
        if x not in ('x', 'y'): raise InputError(f'invalid x: {x!r}, expected "x" or "y".')
        _need0 = True  # flag - still need to find first ax to compare values to.
        axes = self.axs[:, iline] if x=='x' else self.axs[iline, :]
        for ax in axes:
            if ignore_empty and not ax.has_data():
                continue
            ticks_here = ax.get_xticks() if x=='x' else ax.get_yticks()
            _ticklabels = ax.get_xticklabels() if x=='x' else ax.get_yticklabels()
            labels_here = [label.get_text() for label in _ticklabels]
            lims_here = ax.get_xlim() if x=='x' else ax.get_ylim()
            if _need0:
                ticks = ticks_here
                labels = labels_here
                lims = lims_here
                _need0 = False
            else:  # compare to known values
                if not np.array_equal(ticks, ticks_here):
                    return None
                if not all(label==label_here for label, label_here in zip(labels, labels_here)):
                    return None
                if not np.array_equal(lims, lims_here):
                    return None
        if _need0:  # all plots in this row (or column) are empty
            return None
        return (ticks, labels, lims)

    def _remove_redundant_labels(self, x, *, ignore_empty=True):
        '''remove xlabels or ylabels in self which are redundant, i.e. same across col (if 'x') or row (if 'y').
        x: 'x' or 'y'.
        ignore_empty: bool, whether to ignore axes in this column which have no data.
        '''
        if x not in ('x', 'y'): raise InputError(f'invalid x: {x!r}, expected "x" or "y".')
        _override = self.plot_settings['sharexlike'] if x=='x' else self.plot_settings['shareylike']
        for iline in (range(self.ncols) if x=='x' else range(self.nrows)):
            if _override is UNSET:
                common = self._get_common_label(x, iline, ignore_empty=ignore_empty)
            else:
                common = (True if _override else None)
            if common is not None:
                axes = self.axs[:, iline][::-1] if x=='x' else self.axs[iline, :]
                first_nonempty = True
                for ax in axes:
                    if ignore_empty and not ax.has_data():
                        continue
                    elif first_nonempty:  # don't remove labels from first nonempty ax.
                        first_nonempty = False
                    else:
                        ax.set_xlabel('') if x=='x' else ax.set_ylabel('')

    def _get_common_label(self, x, iline, *, ignore_empty=True):
        '''return the x or y label shared by all nonempty plots in this line, or None if that's impossible.
        x: 'x' or 'y'.
        iline: int, index of the line to check. (icol if 'x', irow if 'y')
        ignore_empty: bool, whether to ignore axes in this column which have no data.
        '''
        if x not in ('x', 'y'): raise InputError(f'invalid x: {x!r}, expected "x" or "y".')
        _need0 = True  # flag - still need to find first ax to compare values to.
        axes = self.axs[:, iline] if x=='x' else self.axs[iline, :]
        for ax in axes:
            if ignore_empty and not ax.has_data():
                continue
            label_here = ax.get_xlabel() if x=='x' else ax.get_ylabel()
            if _need0:
                label = label_here
                _need0 = False
            else:  # compare to known values
                if label != label_here:
                    return None
        if _need0:  # all plots in this row (or column) are empty
            return None
        return label

    # # # SUBPLOTS_ADJUST # # #
    @format_docstring(doc=plt.subplots_adjust)
    def subplots_adjust(self, **kw):
        '''plt.subplots_adjust, using values from self.plot_settings by default.
        Note: adjusts the parameters for self.fig, not necessarily plt.gcf().

        plt.subplots_adjust docs:
        -------------------------
        {doc}
        '''
        if self.plot_settings.get('layout') in ('compressed', 'constrained', 'tight'):
            return   # do nothing! subplots_adjust not compatible with this layout.
        kw = self.plot_settings.get_mpl_kwargs('plt.subplots_adjust', **kw)
        return self.fig.subplots_adjust(**kw)

    # # # COLORBAR(S) # # #
    colorbar = staticmethod(colorbar)

    @format_docstring(**_iter_ax_paramdocs)
    @format_docstring(**_paramdocs_colorbar, sub_indent=DEFAULTS.TAB)
    def colorbars_at(self, slice=None, *, sca=False, missing_ok=True, iter_ax='all',
                  location=None, ticks_position=None, pad=None, size=None,
                  kw_add_axes=dict(), **kw_colorbar):
        '''create colorbars for each ax in self.iter_ax(slice), using self.colorbar(...).
        slice: {{slice}}
        sca: bool
            whether to set current axis to the last-created colorbar, after the operation is complete.
            False --> afterwards, plt.gca() will be restored to what it was before this operation began.
        missing_ok: bool
            whether it is okay for some axs to not have a mappable.
        iter_ax: 'all', 'row', or 'col'
            which axes to iterate over.
            slice will be passed to the appropriate iter func. E.g. 'row' --> iter_row(slice, ...)

        The remaining kwargs go to self.colorbar(...):
        {location}
        {ticks_position}
        {pad}
        {size}
        kw_add_axes: dict
            passed to make_cax(..., **kw_add_axes)... which passes it to plt.gcf().add_axes().
        '''
        # [TODO] skip_existing option, to skip axes that already have a colorbar...
        iter_func_lookup = {'all': self.iter_ax, 'row': self.iter_row, 'col': self.iter_col}
        iter_func = iter_func_lookup[iter_ax]
        for _, ax in iter_func(slice, sca=True, restore=(not sca)):
            try:
                cbar = self.colorbar(ax=ax, sca=sca,
                          location=location, ticks_position=ticks_position, pad=pad, size=size,
                          kw_add_axes=kw_add_axes, **kw_colorbar)
            except MappableNotFoundError:
                if missing_ok:
                    pass  # -- we can ignore this error.
                else:
                    raise
            else:
                # formatting
                min_n_ticks = self.plot_settings['min_n_ticks_cbar']
                if min_n_ticks is not None:
                    fail_ok = self.plot_settings['min_n_ticks_cbar_fail_ok']
                    set_min_n_ticks(min_n_ticks, cbar.ax, fail_ok=fail_ok)

    def colorbars_row(self, irow, **kw_colorbars_at):
        '''create colorbars for each ax in row irow, using self.colorbars_at(...).
        Equivalent to self.colorbars_at(irow, iter_ax='row', ...)
        '''
        return self.colorbars_at(irow, iter_ax='row', **kw_colorbars_at)

    def colorbars_col(self, icol, **kw_colorbars_at):
        '''create colorbars for each ax in column icol, using self.colorbars_at(...).
        Equivalent to self.colorbars_at(icol, iter_ax='col', ...)
        '''
        return self.colorbars_at(icol, iter_ax='col', **kw_colorbars_at)

    def colorbars(self, mode='auto', **kw_colorbars_at):
        '''create colorbars for each image in self.
        mode: True, 'auto', 'all', 'row', slice, or tuple
            tells where to create the colorbars.
            True --> use mode='auto'
            'auto' --> infer, by checking which images have equal im.get_clim() and im.cmap.
                       For rows where all image subplots have the same clim and cmap, mode='row'
                       (unless the right-most in row has no image; then use mode='all').
            'all' --> each image gets its own colorbar.
            'row' --> each row gets its own colorbar, at the right-most image in the row.
            slice or tuple --> self.colorbars_at(slice, ...)

            # [TODO] add 'col' option, for horizontal colorbars at top of columns.
            # [TODO] add 'single' option, for one tall colorbar on the right of figure.
            #        [TODO] make_shared_cax() which makes a tall colorbar with same height as multiple axes.
        '''
        if mode is True: mode = 'auto'
        if mode=='auto':
            return self._colorbars_auto(**kw_colorbars_at)
        elif mode=='all':
            return self.colorbars_at(**kw_colorbars_at)
        elif mode=='row':
            return self.colorbars_col(-1, **kw_colorbars_at)
        elif isinstance(mode, (slice, tuple)):
            return self.colorbars_at(mode, **kw_colorbars_at)
        else:
            raise InputError(f'invalid mode: {mode}, expected True, "auto", "all", "row", or a slice or tuple object.')

    def _colorbars_auto(self, **kw_colorbars_at):
        '''create colorbars for each image in self. Equivalent to self.colorbars(mode='auto', ...).
        infer where to put colorbars by checking which images have equal im.get_clim() and im.cmap.
        for rows where all subplots have the same clim and cmap, put colorbar at right.
        '''
        images = self.ax_images()
        row_modes = []
        for row in images:
            first_image = None
            mode = 'all'  # default, unless stuff matches below.
            for i, im in enumerate(row):
                if im is None:
                    if i==len(row)-1:
                        mode = 'all'  # right-most ax has no image --> must use mode='all'
                elif first_image is None:
                    first_image = im  # found the first image!
                elif self.color_scheme_matches(first_image, im):
                    mode = 'row'  # multiple images with same scheme --> can maybe use mode='row'.
                else:
                    mode = 'all'  # not all images in row have same scheme --> use mode='all'.
                    break
            row_modes.append(mode)
        # now, create colorbars according to row_modes.
        for irow, mode in enumerate(row_modes):
            if mode=='all':
                self.colorbars_at((irow, slice(None)), **kw_colorbars_at)
            elif mode=='row':
                self.colorbars_at((irow, -1), **kw_colorbars_at)

    @staticmethod
    def color_scheme_matches(im0, im1):
        '''return whether color for im0 matches color scheme for im1'''
        return im0.get_clim() == im1.get_clim() and im0.cmap == im1.cmap

    # # # MISC FORMATTING # # #
    def apply_misc_formatting(self):
        '''apply misc formatting according to self.plot_settings.'''
        grid = self.plot_settings.get('grid')
        if grid is not UNSET:
            if isinstance(grid, dict):
                self.grid(**grid)
            else:  # bool, hopefully
                self.grid(grid)
        self.set_min_n_ticks()

    def grid(self, *args_grid, **kw_grid):
        '''ax.grid(...) on all axs'''
        for _, ax in self.iter_ax():
            ax.grid(*args_grid, **kw_grid)

    def set_min_n_ticks(self, min_n_ticks=UNSET):
        '''set_min_n_ticks on all axs.
        Use min_n_ticks from self.plot_settings if not provided here.
        '''
        min_n_ticks = self.plot_settings.get('min_n_ticks', min_n_ticks)
        fail_ok = self.plot_settings['min_n_ticks_fail_ok']
        for _, ax in self.iter_ax():
            set_min_n_ticks(min_n_ticks, ax=ax, fail_ok=fail_ok)

    # # # HIDE EMPTY AXES # # #
    def hide_empty_axes(self):
        '''hide axes (ax.set_visible(False)) without any data (check via ax.has_data())'''
        for _, ax in self.iter_ax():
            if not ax.has_data():
                ax.set_visible(False)

    # # # IMAGES # # #
    find_mappable = staticmethod(find_mappable)

    def ax_images(self, *, fill_value=None):
        '''return array of the image on each ax (or fill_value if no image on that ax).'''
        def get_image(ax):
            '''return the image on ax, or None if no image.'''
            try:
                return self.find_mappable(ax=ax)
            except MappableNotFoundError:
                return fill_value
        return self.ax_apply(get_image)

    def iter_images(self):
        '''iterate across axs, yielding ((irow, icol), image) for each ax with an image.
        (axs without image will be skipped.)
        '''
        images = self.ax_images(fill_value=None)
        for (irow, icol), image in np.ndenumerate(images):
            if image is not None:
                yield ((irow, icol), image)

    def ax_cbars(self, *, no_image=None, no_cbar=False):
        '''return array of colorbars associated with images on each ax.
        for ax with no image, value will be `no_image` (default None).
        for ax with image but no colorbar, value will be `no_cbar` (default False).
        '''
        def get_cbar(ax):
            '''return the colorbar on ax, or None if no colorbar.'''
            try:
                result = self.find_mappable(ax=ax)
            except MappableNotFoundError:
                return no_image
            result = getattr(result, 'colorbar', None)
            if result is None:
                return no_cbar
            else:
                return result
        return self.ax_apply(get_cbar)

    def iter_cbars(self):
        '''iterate across axs, yielding ((irow, icol), cbar) for each ax with a colorbar.
        (axs without colorbars will be skipped.)
        '''
        cbars = self.ax_cbars(no_image=None, no_cbar=None)
        for (irow, icol), cbar in np.ndenumerate(cbars):
            if cbar is not None:
                yield ((irow, icol), cbar)

    @property
    def cbars(self):
        '''list of colorbars appearing on any ax across self.
        equivalent: [cbar for ((irow, icol), cbar) in self.iter_cbars()]
        '''
        return [cbar for _, cbar in self.iter_cbars()]
    
    # # # UPDATERS # # #
    def updaters(self, *, fill_value=None):
        '''return array of updaters for each ax (or fill_value if no updater on that ax).'''
        return self.ax_apply(lambda ax: getattr(ax, 'updater', fill_value))

    def set_updater(self, irow, icol, updater):
        '''set updater on ax[irow, icol] to updater'''
        ax = self.axs[irow, icol]
        ax.updater = updater


_dsp = getattr(DEFAULTS.PLOT, 'SUBPLOTS_TYPE', None)
if isinstance(_dsp, type) and issubclass(_dsp, Subplots):
    pass  # don't set DEFAULTS.PLOT in this case; it's already a subclass of Subplots.
else:
    DEFAULTS.PLOT.SUBPLOTS_TYPE = Subplots



"""
File Purpose: xarray attribute accessors.
accessing stuff via xr.DataArray.pc.{attr}, e.g. xr.DataArray.pc.differentiate(...)

# [TODO] __repr__ and help(). E.g. arr.pc.help() gives some help on how to get help,
#   arr.pc.help('') tells all available methods & docs,
#   arr.pc.help('searchstr') tells available methods with 'searchstr' in their name,
#   all similar to the PlasmaCalculator.help() function.
"""
import warnings

import xarray as xr

from ..properties import alias
from ..pytools import help_str
from ..sentinels import UNSET
from ...defaults import DEFAULTS
from ...errors import InputError


''' --------------------- As accessors, on xarrays --------------------- '''
# access via xr.DataArray.pc.{attr}, e.g. xr.DataArray.pc.differentiate(...)

def xarray_register_dataarray_accessor_cls(name, *, warn_if_clsname_match=False):
    '''return class decorator which registers cls as an accessor for xarray.DataArray,
    as per xr.register_dataarray_accessor(name).

    warn_if_clsname_match: bool, default False
        whether to suppress AccessorRegistrationWarning,
            when xr.DataArray.name already exists AND cls.__name__ == xr.DataArray.name.

        This solves the issue that when doing pc=pc.reload(pc),
            (or otherwise using importlib.reload to reload the code attaching the accessor,)
            the code to attach the accessor would be run again, causing a warning.
            But that warning is confusing;
            it's almost certainly just overwriting the name defined by that code originally.
            Using warn_if_clsname_match=False will suppress the warning in that case.
    '''
    def decorated_register_dataaray_accessor(cls):
        with warnings.catch_warnings():  # <-- restore original warnings filter after this.
            if not warn_if_clsname_match:
                try:
                    xr_accessor = getattr(xr.DataArray, name)
                except AttributeError:
                    pass  # that's fine, we'll register cls as an accessor.
                else:
                    if xr_accessor.__name__ == cls.__name__:
                        # suppressing AccessorRegistrationWarning (from xr.register_dataarray_accessor(name))
                        warnings.filterwarnings('ignore', category=xr.core.extensions.AccessorRegistrationWarning)
            return xr.register_dataarray_accessor(name)(cls)
    return decorated_register_dataaray_accessor

# similar to above, but for dataset:
def xarray_register_dataset_accessor_cls(name, *, warn_if_clsname_match=False):
    '''return class decorator which registers cls as an accessor for xarray.Dataset,
    as per xr.register_dataset_accessor(name).

    warn_if_clsname_match: bool, default False
        whether to suppress AccessorRegistrationWarning,
            when xr.Dataset.name already exists AND cls.__name__ == xr.Dataset.name.
    '''
    def decorated_register_dataset_accessor(cls):
        with warnings.catch_warnings():  # <-- restore original warnings filter after this.
            if not warn_if_clsname_match:
                try:
                    xr_accessor = getattr(xr.Dataset, name)
                except AttributeError:
                    pass  # that's fine, we'll register cls as an accessor.
                else:
                    if xr_accessor.__name__ == cls.__name__:
                        # suppressing AccessorRegistrationWarning (from xr.register_dataset_accessor(name))
                        warnings.filterwarnings('ignore', category=xr.core.extensions.AccessorRegistrationWarning)
            return xr.register_dataset_accessor(name)(cls)
    return decorated_register_dataset_accessor


class _BoundObjCaller():
    '''remembers f & instance. calls f(instance.obj, *args, **kw).
    
    Helper class for pcAccessor so that the methods can be compatible with multiprocessing.
    '''
    def __init__(self, f, instance):
        self.f = f
        self.instance = instance
        self.__doc__ = f'caller of {help_str(f, blankline=True)}'
        #functools.update_wrapper(self, f)

    def __call__(self, *args, **kw):   # maybe self, instance, *args, **kw ?
        '''returns self.f(self.instance.obj, *args, **kw).'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return self.f(self.instance.obj, *args, **kw)

    def __repr__(self):
        obj = self.instance.obj
        obj_info = f'{type(obj).__name__} at {hex(id(obj))}'
        f_info = f'{type(self.f).__name__} {self.f.__name__}'
        return f'<{type(self).__name__} of {f_info} for {obj_info}>'


class _ObjCaller():
    '''behaves like a bound method but calls f(self.obj, ...) instead of f(self, ...)'''
    def __init__(self, f):
        self.f = f
        self.__doc__ = f'caller of {help_str(f, blankline=True)}'

    def __get__(self, instance, owner):
        if instance is None:  # called on class, not instance
            return self
        else:
            return _BoundObjCaller(self.f, instance)

    def __repr__(self):
        f_info = f'{type(self.f).__name__} {self.f.__name__}'
        return f'<{type(self).__name__} of {f_info}>'


class XarrayAccessor():
    '''access attributes of DataArrays or Datasets.
    e.g. xr.DataArray.accessor_name.accessor_method(...).

    Not intended for direct use. See pcAccessor for example usage.
    Check cls.registry for details about available methods.
    [TODO] cls.help() which lists available methods...

    accessor_name: None or str
        name for accessor associated with this class.
        E.g., 'pc' for pcAccessor.
        None --> use accessor_name from parent class.
    access_type: UNSET, None, 'array', or 'dataset'
        which type of xarray objects methods registered here apply to (by default).
        UNSET --> use access_type from parent class.
        None --> methods apply to both DataArrays and Datasets.
        'array' --> methods apply to DataArrays only
        'dataset' --> methods apply to Datasets only.

    Implementation assumes when creating a new accessor name, you will follow the pattern:
        class customAccessor(XarrayAccessor, accessor_name='custom', access_type=None): ...
        class customArrayAccessor(customAccessor, access_type='array'): ...
        class customDatasetAccessor(customAccessor, access_type='dataset'): ...
    '''
    def __init_subclass__(cls, *, accessor_name=None, access_type=UNSET, **kwargs):
        super().__init_subclass__(**kwargs)
        if accessor_name is not None:
            cls.accessor_name = accessor_name
        if access_type is None:
            cls.registry = {}  # {name: (type, f)} for all registered methods.
            cls.attrs_registry = {}  # {name: (type, value)} for all registered attrs.
            cls.access_type_to_cls = {}   # {access_type: cls} for single-type access types.
            cls.alias_registry = {}  # {name: [aliases]} for all registered aliases.
        if access_type is not UNSET:
            cls.access_type = access_type
            if not hasattr(cls, 'access_type_to_cls'):
                errmsg = ('access_type_to_cls missing. Probably provided access_type != None, '
                          'for a class which does not inherit from a class with access_type=None.')
                raise NotImplementedError(errmsg)
            if access_type in (None, 'array', 'dataset'):
                if access_type in cls.access_type_to_cls:
                    errmsg = (f'Multiple accessors for access_type={access_type!r}: '
                             f'{cls.access_type_to_cls[access_type].__name__} and {cls.__name__}')
                    raise NotImplementedError(errmsg)
                cls.access_type_to_cls[access_type] = cls
                if access_type == 'array':
                    xarray_register_dataarray_accessor_cls(cls.accessor_name)(cls)
                if access_type == 'dataset':
                    xarray_register_dataset_accessor_cls(cls.accessor_name)(cls)
            else:
                raise InputError(f'Unexpected access_type: {access_type}')

    def __init__(self, xarray_obj):
        self.obj = xarray_obj

    # obj = weakref_property_simple('_obj', doc='''the xarray this object is attached to.''')
    # weakref actually not desireable.. if using it:
    #   e.g. xr.DataArray(0).pc.obj would be None; the array isn't stored so the weakref dies.
    #   meanwhile, arr=xr.DataArray(0); arr.pc.obj would be arr; the weakref didn't die.
    # [TODO][EFF] do xarray data accessors prevent original array from being garbage-collected?

    @classmethod
    def register(cls, f_or_name, *, aliases=[], totype=UNSET, _name=None):
        '''attaches method which applies f(self.obj, *args, **kw) to xr.DataArray.pc.{name}, then returns f.
        (in general, use cls.accessor_name, not necessarily 'pc'; 'pc' is for pcAccessor.)

        This ensures f can be accessed via xr.DataArray.pc.{name} or xr.Dataset.pc.{name}.
            pcAccessor.register --> available on both DataArrays and Datasets.
            pcArrayAccessor.register --> available on DataArrays only.
            pcDatasetAccessor.register --> available on Datasets only.

        f_or_name: str or callable
            str --> returns a function: f -> register(f, name=f_or_name)
            callable --> register this function then return it.
                        will be registered at _name if provided, else at f.__name__.
            This enables this method to be used directly as a decorator, or as a decorator factory.
        aliases: list of str
            aliases for f. Create alias property for each of these.
        totype: UNSET, None, 'array', or 'dataset'
            which type of xarray objects the registered method applies to.
            UNSET --> use cls.access_type.
            None --> methods apply to both DataArrays and Datasets.
            'array' --> methods apply to DataArrays only
            'dataset' --> methods apply to Datasets only.
        _name: str, optional
            name to register f at. If not provided, use f.__name__.
            Not intended to be provided directly.


        Examples (using pcAccessor subclass for concreteness):
            @pcAccessor.register
            def my_method1(xarray_object, arg1):
                print(arg1)
                print(xarray_obj)
            xr.DataArray(data).pc.my_method1(7)   # prints 7 then prints DataArray(data).
            xr.Dataset(data).pc.my_method1(5)   # prints 5 then prints Dataset(data).

            @pcAccessor.register(name='my_method2', totype='array')
            def xarray_my_method2(xarray_object):
                print(xarray_object * 10)
            xr.DataArray(data).pc.my_method2()   # prints 10*DataArray(data).
            xr.Dataset(data).pc.my_method2()   # crashes; my_method2 not registered to datasets.
        '''
        if isinstance(f_or_name, str):
            name = f_or_name
            return lambda f: cls.register(f, _name=name, aliases=aliases, totype=totype)
        else:
            f = f_or_name
            name = _name or f.__name__
            caller = _ObjCaller(f)
            if totype is UNSET:
                totype = cls.access_type
            target_cls = cls.access_type_to_cls[totype]
            setattr(target_cls, name, caller)
            # bookkeeping:
            target_cls.registry[name] = (totype, f)
            # handle aliases if provided:
            for alias_ in aliases:
                setattr(target_cls, alias_, alias(name))
                target_cls.alias_registry.setdefault(name, []).append(alias_)
            return f

    @classmethod
    def register_attr(cls, name, value, *, totype=UNSET):
        '''register cls.{name} = value, using a similar interface as cls.register.
        when totype is UNSET, this is equivalent to:
            cls.{name} = value; cls.attrs_registry[name] = (cls.access_type, value),
        when totype is provided, use cls=cls.access_type_to_cls[totype], instead.

        returns value, after doing setattr(cls, name, value)

        totype: UNSET, None, 'array', or 'dataset'
            which type of xarray objects the registered attr applies to.
            UNSET --> use cls.access_type.
            None --> methods apply to both DataArrays and Datasets.
            'array' --> methods apply to DataArrays only
            'dataset' --> methods apply to Datasets only.

        Examples (using pcAccessor subclass for concreteness):
            pcAccessor.register_attr('MY_CONSTANT1', 7)
            pcAccessor.register_attr('MY_CONSTANT2', 5, totype='array')
            pcAccessor.register_attr('nMbytes', property(lambda self: self.obj.nbytes/1024**2))
            arr = xr.DataArray(some_data)
            ds  = xr.Dataset(other_data)
            arr.pc.MY_CONSTANT1  # == 7
            ds.pc.MY_CONSTANT1   # == 7
            arr.pc.MY_CONSTANT2  # == 5
            ds.pc.MY_CONSTANT2   # crashes; MY_CONSTANT2 not registered to datasets.
            arr.pc.nMbytes       # == DataArray(data).nbytes/1024**2
            ds.pc.nMbytes        # == Dataset(data).nbytes/1024**2
        '''
        if totype is UNSET:
            totype = cls.access_type
        target_cls = cls.access_type_to_cls[totype]
        setattr(target_cls, name, value)
        # bookkeeping:
        target_cls.attrs_registry[name] = (totype, value)
        return value

    registered_methods = property(lambda self: sorted(self.registry.keys()))
    registered_attrs = property(lambda self: sorted(self.attrs_registry.keys()))
    registered_aliases = property(lambda self: {k: v for k, v in sorted(self.alias_registry.items())})

    def __repr__(self):
        contents = []
        methods = self.registered_methods
        if len(methods)>0:
            contents.append(f'registered_methods={methods}')
        attrs = self.registered_attrs
        if len(attrs)>0:
            contents.append(f'registered_attrs={attrs}')
        aliases = self.registered_aliases
        if len(aliases)>0:
            contents.append(f'registered_aliases={aliases}')
        contents_str = ',\n    '.join(contents)
        return f'{type(self).__name__}({object.__repr__(self.obj)},\n    {contents_str})'


class pcAccessor(XarrayAccessor, accessor_name='pc', access_type=None):
    '''access attributes of DataArrays or Datasets e.g. xr.DataArray.pc.differentiate(...).

    This is the base class inherited by pcArrayAccessor and pcDatasetAccessor.
    When deciding where to attach a method, think:
        pcAccessor.register if it applies to DataArrays and Datasets the same way,
        pcAccessor.register(totype='array') if it applies to DataArrays only,
        pcAccessor.register(totype='dataset') if it applies to Datasets only.

    Example:
        @pcAccessor.register
        def my_method(xarray_object, arg1):
            print(arg1)
            print(xarray_obj)
        # --> can later do:
        xr.DataArray(data).pc.my_method(7)   # prints 7 then prints DataArray(data).

        @pcAccessor.register(name='my_method2', totype='array')
        def xarray_my_method2(xarray_object):
            print(xarray_object * 10)
        # --> can later do:
        xr.DataArray(data).pc.my_method2()   # prints DataArray(data).
    '''
    # note: internal pc code should prefer to import the relevant methods from xarray_tools,
    #   for improved compatibility with non-xarray objects.
    # however, scripts should prefer to use the method attached to the array, since it's easier.

    pass  # attach methods & attrs via pcAccessor.register & register_attr.


class pcArrayAccessor(pcAccessor, access_type='array'):
    '''access attributes of DataArrays e.g. xr.DataArray.pc.differentiate(...).
    for more help see help(xr.DataArray.pc).
    '''
    pass  # attach methods & attrs via pcAccessor.register & register_attr with totype='array'


class pcDatasetAccessor(pcAccessor, access_type='dataset'):
    '''access attributes of Datasets e.g. xr.Dataset.pc.differentiate(...).
    for more help see help(xr.Dataset.pc).
    '''
    pass  # attach methods & attrs via pcAccessor.register & register_attr with totype='dataset'


pcAccessor.register_attr('nMbytes', property(lambda self: self.obj.nbytes/1024**2,
                                             doc='''size of array in Mbytes'''))

pcAccessor.register_attr('size', totype='array',
    value=property(lambda self: self.obj.size,
        doc='''total number of elements in the DataArray. Equivalent to array.size.
        Provided for consistent interface for DataArray or Dataset size: use obj.pc.size.'''))

pcAccessor.register_attr('size', totype='dataset',
    value=property(lambda self: sum(v.size for v in self.obj.values()),
        doc='''total number of elements across all values in the Dataset'''))



"""
File Purpose: basic arithmetic (e.g. parenthesis, *, +, log, abs)
"""
import numpy as np
import xarray as xr

from ..quantity_loader import QuantityLoader
from ...tools import np_all_int


class BasicArithmeticLoader(QuantityLoader):
    '''basic arithmetic operations, e.g. *, +, log, abs. For parenthesis, see ParenthesisLoader.
    Also, numerical values (0, 1, and any other integer) are available, for convenience;
        the results for 'N' where N looks like an int, will be an xarray with value int(N).
    '''
    # Caution: the order of these functions matters!
    # Patterns are checked in the order that functions were defined.
    #    (technically - the order of keys in self.KNOWN_PATTERNS... which are tracked via @known_pattern)
    # Examples:
    #    get_plus before get_times --> self('a*b+c') sees '+' first, giving self('a*b')+self('c').
    #    get_minus before get_sqrt --> self('sqrt_a-b') sees '-' first, giving self('sqrt_a')-self('b')

    # # # NUMBERS # # #
    # (numbers come first to ensure, e.g., 1e-7 registers as '1e-7', not '1e'-'7'.
    @known_pattern(r'([+-]?\d+)')  # {int} --> the value of that int.
    def get_int(self, var, *, _match=None):
        '''any integer, as an xarray.'''
        i, = _match.groups()
        return xr.DataArray(int(i))

    @known_var
    def get_0(self):
        '''0, as an xarray. Code can also handle generic ints via self.get_int pattern.'''
        return xr.DataArray(0)

    @known_var
    def get_1(self):
        '''1, as an xarray. Code can also handle generic ints via self.get_int pattern.'''
        return xr.DataArray(1)

    @known_var(aliases=['NaN'])
    def get_nan(self):
        '''NaN, as an xarray.'''
        return xr.DataArray(np.nan)

    @known_var(aliases=['infinity'])
    def get_inf(self):
        '''infinity, as an xarray.'''
        return xr.DataArray(np.inf)

    # non-int numbers
    @known_pattern(r'([+-]?\d+[.]\d+)')  # {float} --> the value of that float.
    def get_float(self, var, *, _match=None):
        '''any float, as an xarray.'''
        f, = _match.groups()
        return xr.DataArray(float(f))

    @known_pattern(r'([+-]?\d+(?:[.]\d+)?)[eE]([+-]?\d+)')  # {number} --> the value of that number.
    def get_sci_number(self, var, *, _match=None):
        '''any number in scientific notation, as an xarray.'''
        base, exp = _match.groups()
        baseval = float(base) if '.' in base else int(base)
        return xr.DataArray(baseval * 10 ** int(exp))

    # # # BASIC OPERATIONS # # #
    @known_pattern(r'(.+)[+](.+)', deps=[0, 1])  # '{var0}+{var1}'
    def get_plus(self, var, *, _match=None):
        '''addition. var0 + var1.'''
        var0, var1 = _match.groups()
        return self(var0) + self(var1)

    @known_pattern(r'(.+)[-](.+)', deps=[0, 1])  # '{var0}-{var1}'
    def get_minus(self, var, *, _match=None):
        '''subtraction. var0 - var1.'''
        var0, var1 = _match.groups()
        return self(var0) - self(var1)

    # note - must be defined after plus & minus, otherwise e.g. '-a+b' --> '-' & 'a+b', <--> '-(a+b)'
    @known_pattern(r'[-](.+)', deps=[0])  # '-{var0}'
    def get_negation(self, var, *, _match=None):
        '''negation. -var.'''
        var, = _match.groups()
        return -self(var)

    @known_pattern(r'(.*[^*])[*]([^*].*)', deps=[0, 1])  # '{var0}*{var1}'. [^*] avoids matching **
    def get_times(self, var, *, _match=None):
        '''multiplication. var0 * var1.'''
        var0, var1 = _match.groups()
        return self(var0) * self(var1)

    @known_pattern(r'(.+)[/÷](.+)', deps=[0, 1])  # '{var0}/{var1}' or '{var0}÷{var1}'
    def get_divide(self, var, *, _match=None):
        '''division. var0 / var1.'''
        var0, var1 = _match.groups()
        return self(var0) / self(var1)

    @known_pattern(r'(.+)[*][*](.+)', deps=[0, 1])  # '{var0}**{var1}'
    def get_power(self, var, *, _match=None):
        '''power. var0 ** var1.'''
        var0, var1 = _match.groups()
        val0 = self(var0)
        val1 = self(var1)
        # caution - avoid naive implementation, because numpy is a bit finicky.
        #    in particular, when result would be complex but val0 & val1 are not,
        #    need to use np.emath.power, instead of **. Otherwise, will get NaN values.
        # need caution unless: any(val0 < 0) and any(val1 not an integer).
        if np.any(val0 < 0) and not np_all_int(val1):
            val0_raised_to_val1 = np.emath.power(val0, val1)
            # sadly, np.emath.power doesn't return an xarray. Make an xarray.
            result, _ = xr.broadcast(val0, val1)
            result = result.copy(data=val0_raised_to_val1)
            return result
        else:
            return val0 ** val1

    # # # BASIC FUNCTIONS # # #
    @known_pattern(r'abs_(.+)', deps=[0])  # 'abs_{var}'
    def get_abs(self, var, *, _match=None):
        '''absolute value. abs(var)'''
        var, = _match.groups()
        return np.abs(self(var))

    @known_pattern(r'log10_(.+)', deps=[0])  # 'log10_{var}'
    def get_log10(self, var, *, _match=None):
        '''log base 10. log10(var)'''
        var, = _match.groups()
        return np.log10(self(var))

    @known_pattern(r'log2_(.+)', deps=[0])  # 'log2_{var}'
    def get_log2(self, var, *, _match=None):
        '''log base 2. log2(var)'''
        var, = _match.groups()
        return np.log2(self(var))

    @known_pattern(r'loge_(.+)', deps=[0])  # 'loge_{var}'  # <-- alias for ln
    @known_pattern(r'ln_(.+)', deps=[0])    # 'ln_{var}'
    def get_ln(self, var, *, _match=None):
        '''log base e. ln(var). ('ln' and 'loge' are aliases. Uses np.log; not np.log10.)'''
        var, = _match.groups()
        return np.log(self(var))

    @known_pattern(r'exp_(.+)', deps=[0])  # 'exp_{var}'
    def get_exp(self, var, *, _match=None):
        '''exponentiation. exp(var). Also known as e^var. See also: get_power'''
        var, = _match.groups()
        return np.exp(self(var))

    @known_pattern(r'sqrt_(.+)', deps=[0])  # 'sqrt_{var}'
    def get_sqrt(self, var, *, _match=None):
        '''square root. sqrt(var)'''
        var, = _match.groups()
        val = self(var)
        result = np.emath.sqrt(val)  # emath --> use complex value, not NaN, if var is negative.
        # sadly, np.emath.sqrt doesn't return an xarray. Make an xarray.
        result = val.copy(data=result)
        return result

    @known_pattern(r'imag_(.+)', deps=[0])  # 'imag_{var}'
    def get_imag(self, var, *, _match=None):
        '''imaginary part. np.imag(self(var))'''
        var, = _match.groups()
        return np.imag(self(var))

    @known_pattern(r'real_(.+)', deps=[0])  # 'real_{var}'
    def get_real(self, var, *, _match=None):
        '''real part. np.real(self(var))'''
        var, = _match.groups()
        return np.real(self(var))



"""
File Purpose: base quantities for CopapicCalculator
"""
import numpy as np

import xarray as xr

from ...errors import (
    InputConflictError, FluidValueError, DimensionAttributeError,
    FormulaMissingError,
)
from ...quantities import AllBasesLoader, SimpleDerivedLoader
from ...tools import UNSET, is_iterable_dim


class CopapicBasesLoader(AllBasesLoader, SimpleDerivedLoader):
    '''base quantities based on Copapic output.'''

    # # # BASE_QUANTS directly from Copapic # # #

    @known_var(load_across_dims=['fluid'])
    def get_m(self):
        '''mass, of a "single particle". For protons, ~= +1 atomic mass unit'''
        return xr.DataArray(self.fluid.m * self.u('M'), attrs=self.units_meta())

    @known_var(load_across_dims=['fluid'])
    def get_q(self):
        '''charge, of a "single particle". For protons, == +1 elementary charge'''
        return xr.DataArray(self.fluid.q * self.u('q'), attrs=self.units_meta())

    @known_var(load_across_dims=['fluid'])
    def get_nusn(self):
        '''collision frequency (of self.fluid) with neutrals.
        "frequency for one particle of species s (self.fluid) to collide with any of the neutrals."
        [TODO] how does this compare with simulated collisions in PIC?
        '''
        return xr.DataArray(self.fluid.params['nu'] * self.u('Hz'), attrs=self.units_meta())

    @known_var(load_across_dims=['component'], aliases=['dspace'])
    def get_ds(self):
        '''grid spacing (of output files). vector(ds), e.g. [dx, dy, dz]. Depends on self.component.
        '''
        x = str(self.component)
        dx_raw = self.input_deck.get_dspace(x) 
        dx = dx_raw * self.u('length')
        return xr.DataArray(dx, attrs=self.units_meta())

    @known_var(dims=['snap', 'fluid'], aliases=['den'], deps=['n'])
    def get_deltafrac_n(self):
        '''normalized density perturbation. (directly from Copapic.) deltafrac_n = (n - n0) / n0.'''
        n = self('n')
        n0 = self('n0')
        return (n - n0) / n0

    @known_var(load_across_dims=['fluid'])
    def get_n0(self):
        '''background density. (directly from Copapic.)
        '''
        try:
            n0 = self.fluid.n0
        except AttributeError:
            errmsg = f'{type(self.fluid).__name__} object has no attribute "n0".'
            raise DimensionAttributeError(errmsg) from None
        return xr.DataArray(n0 * self.u('n'), attrs=self.units_meta())
    
    @known_var(dims=['snap'])
    def get_phi(self):
        '''electric potential. (directly from Copapic)'''
        return self.load_maindims_var_across_dims('phi', u='E length', dims=['snap'])

    @known_var(load_across_dims=['component'])
    def get_E_ext(self):
        '''external electric field. (directly from Copapic).'''
        x = str(self.component)
        E = self.input_deck.get('E_extern', [0, 0, 0])
        if x == 'x':
            Ex = E[0] * self.u('E')
        elif x =='y':
            Ex = E[1] * self.u('E')
        elif x == 'z':
            Ex = E[2] * self.u('E')
        return xr.DataArray(Ex, attrs=self.units_meta())

    @known_var(load_across_dims=['component'])
    def get_B(self):
        '''magnetic field. (directly from Copapic)'''
        x = str(self.component)
        B = self.input_deck.get('B_extern', [0, 0, 0])
        if x == 'x':
            Bx = B[0] * self.u('b')
        elif x == 'y':
            Bx = B[1] * self.u('b')
        elif x == 'z':
            Bx = B[2] * self.u('b')
        return xr.DataArray(Bx, attrs=self.units_meta())

    @known_var(deps=['E_ext', 'E_phi'])
    def get_E(self):
        '''electric field. E = E_external + E_phi = E_external - grad(phi)'''
        E_ext = self('E_ext')
        E_phi = self('E_phi')
        return E_ext + E_phi
    

    @known_var(deps=['grad_phi'])
    def get_E_phi(self):
        '''electric field, from phi. E_phi = -grad(phi). Doesn't include E_ext component.'''
        return -self('grad_phi')
    

    @known_var(dims=['snap', 'fluid'])
    def get_n(self):
        '''Number density. Directly from Copapic'''
        return self.load_maindims_var_across_dims('den', dims=['snap', 'fluid'])  # [dimensionless]

    @known_var(load_across_dims=['fluid'])
    def get_collType(self):
        '''collision type. (directly from Copapic)'''
        return xr.DataArray(self.fluid.params["collision_type"])
    
    @known_var(load_across_dims=['fluid'])
    def get_vadvanceType(self):
        '''advance type. (directly from Copapic)'''
        return xr.DataArray(self.fluid.params["vadvance_type"])
    
    @known_var(load_across_dims=['fluid'])
    def get_npercell(self):
        '''number of particles per cell. (directly from Copapic)'''
        return xr.DataArray(self.fluid.params["npercell"], attrs=self.units_meta())
    
    @known_var(load_across_dims=['fluid', 'component'])
    def get_v0(self):
        '''Initial Drift Velocity. (directly from Copapic)'''
        x = str(self.component)
        v0 = self.fluid.params["v0"]
        if x == 'x':
            v0 = v0[0] * self.u('l') / self.u('t')
        elif x == 'y':
            v0 = v0[1] * self.u('l') / self.u('t')
        elif x == 'z':
            v0 = v0[2] * self.u('l') / self.u('t')
        return xr.DataArray(v0, attrs=self.units_meta())
    
    @known_var(load_across_dims=['fluid', 'component'])
    def get_vth(self):
        '''thermal velocity. (directly from Copapic)'''
        x = str(self.component)
        vth = self.fluid.params["vth"]
        if x == 'x':
            vth = vth[0] * self.u('l') / self.u('t')
        elif x == 'y':
            vth = vth[1] * self.u('l') / self.u('t')
        elif x == 'z':
            vth = vth[2] * self.u('l') / self.u('t')
        return xr.DataArray(vth, attrs=self.units_meta())
    
    @known_var(load_across_dims=['component','snap'])
    def get_flux(self):
        '''flux. (directly from Copapic)'''
        x = str(self.component)
        varname = f'flux_{x}'
        return self.load_maindims_var_across_dims(varname, dims=['snap'])  # [dimensionless]
    
    
    @known_var(load_across_dims=['component','snap'])
    def get_Eraw(self):
        '''Electric field. (directly from Copapic)'''
        x = str(self.component)
        varname = f'E_{x}'
        return self.load_maindims_var_across_dims(varname, u='E', dims=['snap'])  # [dimensionless]
    
    @known_var(load_across_dims=['snap'])
    def get_rho(self):
        '''charge density. (directly from Copapic)'''
        return self.load_maindims_var_across_dims('rho', dims=['snap'])


"""
File Purpose: Miscellaneous quality-of-life functions for Object Oriented Programming tasks.
"""

from ...defaults import DEFAULTS


''' --------------------- Apply if attribute exists --------------------- '''

def apply(x, fstr, *args, **kwargs):
    '''return x.fstr(*args, **kwargs), or x if x doesn't have an 'fstr' attribute.'''
    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
    # pop default if it was provided.
    doing_default = 'default' in kwargs
    if doing_default:
        default = kwargs.pop('default')
    # call x.fstr(*args, **kwargs)   # (kwargs with 'default' popped.)
    if hasattr(x, fstr):
        return getattr(x, fstr)(*args, **kwargs)
    elif doing_default:
        return default
    else:
        return x


''' --------------------- Metaclasses --------------------- '''

class MetaClsRepr(type):
    '''metaclass which affects repr for classes.
    classes can use metaclass=MetaClsRepr and define classmethod __cls_repr__ method,
        to define their own repr.
    '''
    def __repr__(cls):
        if hasattr(cls, '__cls_repr__'):
            return cls.__cls_repr__()
        else:
            return super().__repr__()



"""
File Purpose: testing tools from tools.xarray_tools.xarray_sci
"""
import pytest
import numpy as np
import xarray as xr

import PlasmaCalcs as pc


def test_xarray_interp_inverse():
    # silly little example, to make it easy to understand & follow:
    # arr looks like: [[0, 10, ..., 40], [1, 11, ..., 41], ..., [9, 19, ..., 49]].
    tt = pc.xrrange(5, 'tens')
    oo = pc.xrrange(10, 'ones')
    arr = (10 * tt + oo).rename('n')
    assert pc.xarray_interp_inverse(arr, tens=1, n=17) == 7  # n=17, tens=1 --> ones=7
    assert arr.pc.interp_inverse(tens=1, n=17) == 7  # more convenient syntax works too.
    # more exact matches:
    assert arr.pc.interp_inverse(tens=2, n=23) == 3  # n=23, tens=2 --> ones=3
    assert arr.pc.interp_inverse(n=20, tens=2) == 0  # n=20, tens=2 --> ones=0
    assert arr.pc.interp_inverse(n=18, ones=8) == 1  # n=18, ones=8 --> tens=1
    assert arr.pc.interp_inverse(n=29, ones=9) == 2  # n=29, ones=9 --> tens=2
    # inexact matches:
    assert arr.pc.interp_inverse(n=15, ones=0) == 1.5  # ones=0 --> tens needs 0.5
    assert arr.pc.interp_inverse(n=27, ones=3) == 2.4
    assert arr.pc.interp_inverse(n=23.5, tens=2) == 3.5
    assert arr.pc.interp_inverse(n=27, tens=2.5) == 2
    # unreachable array values (n) gives nan
    assert np.isnan(arr.pc.interp_inverse(n=27, tens=3))  # ones can't compensate
    # out of bounds input coords (ones or tens) raises DimensionValueError (regardless of n)
    with pytest.raises(pc.DimensionValueError):
        arr.pc.interp_inverse(n=27, ones=10)
    with pytest.raises(pc.DimensionValueError):
        arr.pc.interp_inverse(n=27.5, ones=-1)
    with pytest.raises(pc.DimensionValueError):
        arr.pc.interp_inverse(n=None, ones=-1)
    with pytest.raises(pc.DimensionValueError):
        arr.pc.interp_inverse(n=-1000, tens=5)
    with pytest.raises(pc.DimensionValueError):
        arr.pc.interp_inverse(n=10000, tens=-1)
    with pytest.raises(pc.DimensionValueError):
        arr.pc.interp_inverse(n=None, tens=-1)
    # coords as arrays:
    assert np.all(arr.pc.interp_inverse(ones=[7, 2], n=27) == [2.0, 2.5])
    vals = arr.pc.interp_inverse(tens=[2,3], n=27)
    assert vals[0] == 7
    assert np.isnan(vals[1])
    # array.name var as array:
    assert np.all(arr.pc.interp_inverse(ones=5, n=[20, 25, 30, 36]) == [1.5, 2, 2.5, 3.1])
    # both as array:
    vals = arr.pc.interp_inverse(ones=[0, 7, 9], n=[27, 29])
    assert np.all(vals.isel(n=0) == [2.7, 2. , 1.8])
    assert np.all(vals.isel(n=1) == [2.9, 2.2, 2. ])
    # interp_inverse + interp --> original values
    result = arr.pc.interp_inverse(ones=[0, 7, 9], n=[27, 29])
    orig = arr.interp(ones=[0, 7, 9], tens=result)
    assert np.allclose(orig.isel(ones=0), [27, 29])
    assert np.allclose(orig.isel(ones=1), [27, 29])
    assert np.allclose(orig.isel(ones=2), [27, 29])
    assert np.allclose(orig, [27, 29])  # broadcasting is fine^^
    # [TODO] could probably do more checks,
    #   e.g. what happens if input array is not sorted,
    #   or what happens if it does not have only unique values.
    # [TODO] check coords & dims handled appropriately when inputs are DataArrays,
    #   especially DataArrays with different dimensions than the coord & array names.

def test_xarray_map():
    xx = pc.xrrange(4, 'x')
    yy = pc.xrrange(3, 'y')
    arr = -xx * yy + yy
    assert np.all(arr == [[0, 1, 2], [0, 0, 0], [0, -1, -2], [0, -2, -4]])
    sumx = arr.pc.map(np.sum, axis='x')
    assert np.all(sumx == [0, -2, -4])
    sumy = arr.pc.map(np.sum, axis='y')
    assert np.all(sumy == [3, 0, -3, -6])
    argsortx = arr.pc.map(np.argsort, axis='x')
    assert np.all(argsortx == [[0, 3, 3], [1, 2, 2], [2, 1, 1], [3, 0, 0]])
    argsorty = arr.pc.map(np.argsort, axis='y')
    assert np.all(argsorty == [[0, 1, 2], [0, 1, 2], [2, 1, 0], [2, 1, 0]])
